from __future__ import annotations

import os
import subprocess
import sys
import tkinter as tk
from pathlib import Path
from tkinter import messagebox, ttk

from ...config import host_id, init_config, open_scope
from ..aurelia_theme import get_palette, use


def _launch(path: Path) -> None:  # pragma: no cover - best effort
    try:
        if sys.platform.startswith("win"):
            os.startfile(path)  # type: ignore[attr-defined]
        elif sys.platform == "darwin":
            subprocess.run(["open", str(path)], check=False)
        else:
            subprocess.run(["xdg-open", str(path)], check=False)
    except Exception:
        pass


def launch() -> None:  # pragma: no cover - GUI interactions
    root = tk.Tk()
    use(root)
    palette = get_palette()
    root.title("Sigil Config")
    root.configure(bg=palette["bg"])  # type: ignore[call-arg]
    ttk.Label(root, text=f"Host: {host_id()}", style="Card.TLabel").pack(padx=10, pady=10)

    def do_open() -> None:
        path = open_scope("user-custom", "user")
        try:
            _launch(path)
        except Exception:
            try:
                messagebox.showerror("Sigil Config", f"Could not open {path}")
            except Exception:
                pass

    def do_init() -> None:
        path = init_config("user-custom", "user")
        try:
            messagebox.showinfo("Sigil Config", f"Initialized user-custom at {path}")
        except Exception:
            pass
        do_open()

    ttk.Button(root, text="Initialize User Custom", command=do_init).pack(pady=5)
    ttk.Button(root, text="Open user config folder", command=do_open).pack(pady=5)

    root.mainloop()



"""Reusable tkinter widgets used by :mod:`pysigil.ui.tk`.

This module provides small standalone widgets used by the tkinter view
layer.  It ports components originally prototyped during the UI design
process, namely :class:`HoverTip` for delayed hover tooltips and
:class:`PillButton`, a rounded button displaying the value state of a
setting.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, Literal

from ..aurelia_theme import get_palette

try:  # pragma: no cover - importing tkinter is environment dependent
    import tkinter as tk
    import tkinter.font as tkfont
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore
    tkfont = None  # type: ignore


# Maximum width of tooltip labels before wrapping (pixels)
_TIP_WRAP_LENGTH = 480


class HoverTip:
    """Display a tooltip when hovering over a widget."""

    def __init__(
        self,
        widget: tk.Widget,
        text_fn: Callable[[], str | None],
        *,
        delay: int = 250,
    ) -> None:
        self.widget = widget
        self.text_fn = text_fn
        self.delay = delay
        self._after: str | None = None
        self.tip: tk.Toplevel | None = None
        widget.bind("<Enter>", self._schedule, add=True)
        widget.bind("<Leave>", self._hide, add=True)

    def _schedule(self, _event: tk.Event) -> None:
        self._after = self.widget.after(self.delay, self._show)

    def _show(self) -> None:
        if self.tip:
            return
        txt = self.text_fn() or ""
        if not txt:
            return
        x = self.widget.winfo_rootx() + 4
        y = self.widget.winfo_rooty() + self.widget.winfo_height() + 6
        self.tip = tk.Toplevel(self.widget)
        self.tip.wm_overrideredirect(True)
        self.tip.attributes("-topmost", True)
        self.tip.wm_geometry(f"+{x}+{y}")
        palette = get_palette()
        bg = palette["tooltip_bg"]
        fg = palette["tooltip_fg"]
        frame = tk.Frame(self.tip, bg=bg, bd=0)
        frame.pack()
        label = tk.Label(
            frame,
            text=txt,
            bg=bg,
            fg=fg,
            padx=8,
            pady=6,
            justify="left",
            wraplength=_TIP_WRAP_LENGTH,
        )
        label.pack()

    def _hide(self, _event: tk.Event) -> None:
        if self._after:
            self.widget.after_cancel(self._after)
            self._after = None
        if self.tip:
            tip = self.tip
            self.tip = None
            tip.after(120, tip.destroy)


class PillButton(tk.Canvas):
    """Rounded button indicating value state.

    ``state`` can be one of ``"effective"``, ``"present"``, ``"empty"``,
    ``"disabled"`` or ``"synthetic"``.
    """

    def __init__(
        self,
        master: tk.Widget,
        *,
        text: str,
        color: str,
        state: Literal["effective", "present", "empty", "disabled", "synthetic"],
        value_provider: Callable[[], Any],
        clickable: bool = True,
        on_click: Callable[[], None] | None = None,
        tooltip_title: str | None = None,
        tooltip_desc: str | None = None,
        locked: bool = False,
    ) -> None:
        super().__init__(master, height=28, highlightthickness=0, bd=0)
        self.text = text
        self.tooltip_title = tooltip_title or text
        self.tooltip_desc = tooltip_desc
        self.color = color
        self.state = state
        self.locked = locked
        self.clickable = clickable
        self.on_click = on_click
        self.font = tkfont.Font(size=9, weight="bold") if tkfont else None
        self.value_provider = value_provider
        self.pad_x = 14
        self.rad = 14
        self.bind("<Configure>", lambda e: self._draw())
        if self.clickable:
            self.bind("<Button-1>", lambda e: on_click() if on_click else None)
            self.configure(cursor="hand2")
        else:
            self.configure(cursor="arrow")
        self.configure(takefocus=1)
        self.bind("<FocusIn>", lambda e: self._draw())
        self.bind("<FocusOut>", lambda e: self._draw())
        self.bind(
            "<Key-Return>",
            lambda e: self.on_click() if self.clickable and self.on_click else None,
        )
        self.bind(
            "<Key-space>",
            lambda e: self.on_click() if self.clickable and self.on_click else None,
        )
        HoverTip(self, self._tip_text)
        self._draw(initial=True)

    # -- drawing ------------------------------------------------------------

    def _measure_width(self) -> int:
        if self.font is None:
            return 64
        text_w = self.font.measure(self.text)
        return max(64, text_w + self.pad_x * 2)

    def _draw(self, initial: bool = False) -> None:
        w = self._measure_width()
        self.configure(width=w)
        self.delete("all")
        palette = get_palette()
        self.configure(bg=palette["card"])
        card = palette["card"]
        card_edge = palette["card_edge"]
        ink_muted = palette["ink_muted"]
        if self.state == "effective":
            fill = outline = self.color
            fg = palette["on_primary"]
            border_w = 1
        elif self.state == "present":
            fill = card
            outline = self.color
            fg = self.color
            border_w = 2
        elif self.state == "disabled":
            fill = card
            outline = card_edge
            fg = ink_muted
            border_w = 1
        elif self.state == "synthetic":
            fill = card
            outline = self.color
            fg = self.color
            border_w = 1
        else:  # empty
            fill = card
            outline = card_edge
            fg = ink_muted
            border_w = 1
        r = self.rad
        self._round_rect(1, 1, w - 1, 26, r, fill=fill, outline=outline, width=border_w)
        self.create_text(w / 2, 14, text=self.text, fill=fg, font=self.font)
        if self.focus_displayof() is self:
            self.create_rectangle(
                3,
                3,
                w - 3,
                24,
                outline=palette["ink"],
                dash=(2, 2),
            )

    def _round_rect(self, x1: int, y1: int, x2: int, y2: int, r: int, **kw: Any) -> int:
        pts = [
            x1 + r,
            y1,
            x2 - r,
            y1,
            x2,
            y1,
            x2,
            y1 + r,
            x2,
            y2 - r,
            x2,
            y2,
            x2 - r,
            y2,
            x1 + r,
            y2,
            x1,
            y2,
            x1,
            y2 - r,
            x1,
            y1 + r,
            x1,
            y1,
        ]
        return self.create_polygon(pts, smooth=True, **kw)

    def _tip_text(self) -> str:
        val = self.value_provider()
        val_txt = str(val) if val is not None else "—"
        parts = [self.tooltip_title]
        parts.append(f"Value: {val_txt}")
        if self.tooltip_desc:
            parts.append('________________________________')
            parts.append(self.tooltip_desc)

        return "\n".join(parts)


__all__ = [
    "HoverTip",
    "PillButton",
]



"""High level orchestration layer for Sigil configuration.

This module exposes a thin façade that ties together a specification
backend (storing provider metadata) and the existing configuration
backend.  It offers a small, UI agnostic API for managing provider
metadata and reading/writing typed configuration values.  The
implementation is deliberately compact but mirrors the design proposed in
the accompanying user request so that it can grow organically.
"""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace
from pathlib import Path
from typing import Literal

from .authoring import get as _get_dev_link, normalize_provider_id
from .errors import (
    DevLinkNotFoundError,
    DuplicateFieldError,
    PolicyError,
    SigilLoadError,
    UnknownFieldError,
    UnknownProviderError,
    ValidationError,
)
from .root import ProjectRootNotFoundError
from .settings_metadata import (
    TYPE_REGISTRY,
    FieldSpec,
    FieldValue,
    IniFileBackend,
    IniSpecBackend,
    ProviderManager,
    ProviderSpec,
    SigilBackend,
    SpecBackend,
    save_provider_spec,
)


@dataclass
class AuthorContext:
    provider_id: str
    dev_root: Path
    mode: Literal["bootstrap", "edit"]
    spec: ProviderSpec | None = None

# ---------------------------------------------------------------------------
# orchestrator implementation
# ---------------------------------------------------------------------------


class Orchestrator:
    """High level façade coordinating spec and config backends."""

    def __init__(
        self,
        spec_backend: SpecBackend | None = None,
        config_backend: SigilBackend | None = None,
    ) -> None:
        if spec_backend is None:
            spec_backend = IniSpecBackend()
        if config_backend is None:
            config_backend = IniFileBackend()
        self.spec_backend = spec_backend
        self.config_backend = config_backend

    # ---- Development links ----
    def has_dev_link(self, provider_id: str) -> bool:
        pid = normalize_provider_id(provider_id)
        return _get_dev_link(pid) is not None

    def get_dev_link(self, provider_id: str) -> Path:
        pid = normalize_provider_id(provider_id)
        dl = _get_dev_link(pid)
        if dl is None:
            raise DevLinkNotFoundError(pid)
        return dl.defaults_path.parent.parent

    def spec_exists(self, provider_id: str) -> bool:
        pid = normalize_provider_id(provider_id)
        return self.spec_backend.exists(pid)

    def load_author_context(self, provider_id: str) -> AuthorContext:
        pid = normalize_provider_id(provider_id)
        dl = _get_dev_link(pid)
        if dl is None:
            raise DevLinkNotFoundError(pid)
        dev_root = dl.defaults_path.parent.parent
        if self.spec_backend.exists(pid):
            spec = self.spec_backend.get_spec(pid)
            return AuthorContext(provider_id=pid, dev_root=dev_root, mode="edit", spec=spec)
        return AuthorContext(provider_id=pid, dev_root=dev_root, mode="bootstrap", spec=None)

    # ---- Provider / package metadata ----
    def register_provider(
        self,
        provider_id: str,
        *,
        title: str | None = None,
        description: str | None = None,
    ) -> ProviderSpec:
        """Create and store a new provider specification.

        Raises
        ------
        DuplicateProviderError
            If the provider already exists.
        PolicyError
            If the provider metadata cannot be written due to project
            configuration.
        """
        pid = normalize_provider_id(provider_id)
        spec = ProviderSpec(
            provider_id=pid,
            schema_version="0",
            title=title,
            description=description,
        )
        try:
            self.spec_backend.create_spec(spec)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc
        return spec

    def edit_provider(
        self,
        provider_id: str,
        *,
        title: str | None = None,
        description: str | None = None,
    ) -> ProviderSpec:
        """Modify metadata for an existing provider.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        PolicyError
            If the provider metadata cannot be written due to project
            configuration.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        etag = self.spec_backend.etag(pid)
        updated = replace(
            spec,
            title=spec.title if title is None else title,
            description=spec.description if description is None else description,
        )
        try:
            self.spec_backend.save_spec(updated, expected_etag=etag)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc
        return updated

    def delete_provider(self, provider_id: str) -> None:
        """Delete a provider specification.

        Raises
        ------
        PolicyError
            If the provider metadata cannot be removed due to project
            configuration.
        """
        pid = normalize_provider_id(provider_id)
        try:
            self.spec_backend.delete_spec(pid)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

    # ---- Field metadata ----
    def add_field(
        self,
        provider_id: str,
        *,
        key: str,
        type: str,
        label: str | None = None,
        description_short: str | None = None,
        description: str | None = None,
        options: Mapping[str, object] | None = None,
        section: str | None = None,
        order: int | None = None,
    ) -> FieldSpec:
        """Add a new field to a provider specification.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        DuplicateFieldError
            If a field with *key* already exists.
        PolicyError
            If the provider metadata cannot be written due to project
            configuration.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        if key in {f.key for f in spec.fields}:
            raise DuplicateFieldError(key)
        field = FieldSpec(
            key=key,
            type=type,
            label=label,
            description_short=description_short,
            description=description,
            options=dict(options) if options is not None else {},
            section=section,
            order=order,
        )
        new_spec = replace(spec, fields=tuple(spec.fields) + (field,))
        etag = self.spec_backend.etag(pid)
        try:
            self.spec_backend.save_spec(new_spec, expected_etag=etag)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc
        mgr = ProviderManager(new_spec, self.config_backend)
        try:  # ensure default section exists for user scope
            mgr.init("user")
        except ProjectRootNotFoundError as exc:  # pragma: no cover - defensive
            raise PolicyError(str(exc)) from exc
        return field

    def edit_field(
        self,
        provider_id: str,
        key: str,
        *,
        new_key: str | None = None,
        new_type: str | None = None,
        label: str | None = None,
        description_short: str | None = None,
        description: str | None = None,
        options: Mapping[str, object] | None = None,
        section: str | None = None,
        order: int | None = None,
        on_type_change: Literal["convert", "clear"] = "convert",
    ) -> FieldSpec:
        """Modify an existing field definition.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        UnknownFieldError
            If *key* does not exist.
        DuplicateFieldError
            If *new_key* conflicts with another field.
        PolicyError
            If metadata or configuration cannot be written.
        SigilLoadError
            If existing configuration values cannot be read.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        fields = list(spec.fields)
        try:
            index = next(i for i, f in enumerate(fields) if f.key == key)
        except StopIteration as exc:
            raise UnknownFieldError(key) from exc

        old_field = fields[index]
        nk = new_key or key
        nt = new_type or old_field.type

        if nk != key and nk in {f.key for f in fields}:
            raise DuplicateFieldError(nk)

        fields[index] = FieldSpec(
            key=nk,
            type=nt,
            label=old_field.label if label is None else label,
            description_short=(
                old_field.description_short
                if description_short is None
                else description_short
            ),
            description=old_field.description if description is None else description,
            options=old_field.options if options is None else dict(options),
            section=old_field.section if section is None else section,
            order=old_field.order if order is None else order,
        )

        raw_map, source_map = self.config_backend.read_merged(pid)
        raw = raw_map.get(key)
        source = source_map.get(key)
        if raw is not None and source is not None:
            scope = source.split("-")[0]
            target = self.config_backend.write_target_for(pid)
            new_raw = raw
            if nt != old_field.type:
                if on_type_change == "convert":
                    adapter = TYPE_REGISTRY[nt].adapter
                    try:
                        value = adapter.parse(raw)
                        new_raw = adapter.serialize(value)
                    except Exception as exc:
                        raise ValidationError(str(exc)) from exc
                elif on_type_change == "clear":
                    new_raw = None
            if nk != key:
                if new_raw is not None:
                    self.config_backend.write_key(
                        pid, nk, new_raw, scope=scope, target_kind=target
                    )
                else:
                    self.config_backend.remove_key(
                        pid, nk, scope=scope, target_kind=target
                    )
                self.config_backend.remove_key(
                    pid, key, scope=scope, target_kind=target
                )
            else:
                if new_raw is None:
                    self.config_backend.remove_key(
                        pid, nk, scope=scope, target_kind=target
                    )
                elif new_raw != raw:
                    self.config_backend.write_key(
                        pid, nk, new_raw, scope=scope, target_kind=target
                    )

        new_spec = replace(spec, fields=tuple(fields))
        etag = self.spec_backend.etag(pid)
        try:
            self.spec_backend.save_spec(new_spec, expected_etag=etag)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

        return fields[index]

    def delete_field(
        self,
        provider_id: str,
        key: str,
        *,
        remove_values: bool = False,
        scopes: tuple[str, ...] = ("user", "project"),
    ) -> None:
        """Remove a field from a provider specification.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        UnknownFieldError
            If *key* does not exist.
        PolicyError
            If metadata or configuration cannot be written.
        SigilLoadError
            If existing configuration values cannot be read when
            removing stored values.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        fields = [f for f in spec.fields if f.key != key]
        if len(fields) == len(spec.fields):
            raise UnknownFieldError(key)
        new_spec = replace(spec, fields=tuple(fields))
        etag = self.spec_backend.etag(pid)
        try:
            self.spec_backend.save_spec(new_spec, expected_etag=etag)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

        if remove_values:
            target = self.config_backend.write_target_for(pid)
            for scope in scopes:
                try:
                    self.config_backend.remove_key(
                        pid, key, scope=scope, target_kind=target
                    )
                except ProjectRootNotFoundError as exc:  # pragma: no cover - defensive
                    raise PolicyError(str(exc)) from exc

    def get_sections_order(self, provider_id: str) -> list[str] | None:
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        return list(spec.sections_order) if spec.sections_order is not None else None

    def set_sections_order(self, provider_id: str, seq: list[str]) -> None:
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        new_spec = replace(spec, sections_order=tuple(seq))
        etag = self.spec_backend.etag(pid)
        try:
            self.spec_backend.save_spec(new_spec, expected_etag=etag)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

    def get_sections_collapsed(self, provider_id: str) -> list[str] | None:
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        return (
            list(spec.sections_collapsed)
            if spec.sections_collapsed is not None
            else None
        )

    def set_sections_collapsed(self, provider_id: str, seq: list[str]) -> None:
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        new_spec = replace(spec, sections_collapsed=tuple(seq))
        etag = self.spec_backend.etag(pid)
        try:
            self.spec_backend.save_spec(new_spec, expected_etag=etag)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

    def patch_fields(self, provider_id: str, updates: list[dict[str, object]]) -> None:
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        fields = list(spec.fields)
        index = {f.key: i for i, f in enumerate(fields)}
        for upd in updates:
            key = str(upd["key"])
            idx = index.get(key)
            if idx is None:
                continue
            field = fields[idx]
            fields[idx] = FieldSpec(
                key=field.key,
                type=field.type,
                label=field.label,
                description_short=field.description_short,
                description=field.description,
                options=field.options,
                section=upd.get("section", field.section),
                order=upd.get("order", field.order),
            )
        new_spec = replace(spec, fields=tuple(fields))
        etag = self.spec_backend.etag(pid)
        try:
            self.spec_backend.save_spec(new_spec, expected_etag=etag)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

    # ---- Discovery ----
    def list_providers(self) -> list[str]:
        """Return identifiers for all registered providers."""
        return self.spec_backend.get_provider_ids()

    def list_fields(self, provider_id: str) -> list[FieldSpec]:
        """List field specifications for *provider_id*.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        return list(spec.fields)

    def find_untracked_keys(self, provider_id: str) -> list[str]:
        """Return config keys not described in provider metadata.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        SigilLoadError
            If configuration values cannot be read.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        raw_map, _ = self.config_backend.read_merged(pid)
        tracked = {f.key for f in spec.fields}
        return sorted(set(raw_map) - tracked)

    def adopt_untracked(self, provider_id: str, mapping: dict[str, str]) -> list[FieldSpec]:
        """Create field specs for previously untracked keys.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        DuplicateFieldError
            If any key already exists.
        PolicyError
            If metadata cannot be written.
        """
        added: list[FieldSpec] = []
        for key, type in mapping.items():
            added.append(self.add_field(provider_id, key=key, type=type))
        return added

    # ---- Values ----
    def _manager(self, provider_id: str) -> ProviderManager:
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        return ProviderManager(spec, self.config_backend)

    def get_effective(self, provider_id: str) -> dict[str, FieldValue]:
        """Return effective configuration values for *provider_id*.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        SigilLoadError
            If configuration values cannot be read.
        """
        return self._manager(provider_id).effective()

    def get_layers(self, provider_id: str) -> dict[str, dict[str, FieldValue | None]]:
        """Return values for all scopes for *provider_id*."""
        return self._manager(provider_id).layers()

    def set_value(
        self,
        provider_id: str,
        key: str,
        value: object,
        *,
        scope: Literal["user", "project", "default"] = "user",
    ) -> None:
        """Persist *value* for *key*.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        UnknownFieldError
            If *key* is not defined.
        ValidationError
            If *value* fails validation.
        PolicyError
            If configuration cannot be written due to project settings.
        SigilLoadError
            If configuration values cannot be read.
        """
        mgr = self._manager(provider_id)
        try:
            mgr.set(key, value, scope=scope)
        except (TypeError, ValueError) as exc:
            raise ValidationError(str(exc)) from exc
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

    def clear_value(
        self,
        provider_id: str,
        key: str,
        *,
        scope: Literal["user", "project", "default"] = "user",
    ) -> None:
        """Remove the stored value for *key*.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        UnknownFieldError
            If *key* is not defined.
        PolicyError
            If configuration cannot be written.
        SigilLoadError
            If configuration values cannot be read.
        """
        mgr = self._manager(provider_id)
        try:
            mgr.clear(key, scope=scope)
        except ProjectRootNotFoundError as exc:
            raise PolicyError(str(exc)) from exc

    def set_many(
        self,
        provider_id: str,
        updates: dict[str, object],
        *,
        scope: Literal["user", "project", "default"] = "user",
        atomic: bool = True,
    ) -> None:

        """Set multiple configuration values.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        UnknownFieldError
            If any key is not defined.
        ValidationError
            If any value fails validation.
        PolicyError
            If configuration cannot be written.
        SigilLoadError
            If configuration values cannot be read.
Set multiple values for *provider_id* at once.

        When ``atomic`` is true (the default) all updates are validated and
        written as a single transaction.  If validation or writing fails no
        changes are persisted.

        """
        mgr = self._manager(provider_id)
        if atomic:
            fields = mgr._fields  # pylint: disable=protected-access
            for key, value in updates.items():
                field = fields.get(key)
                if field is None:
                    raise UnknownFieldError(key)
                adapter = TYPE_REGISTRY[field.type].adapter
                try:
                    adapter.validate(value, field)
                except (TypeError, ValueError) as exc:
                    raise ValidationError(str(exc)) from exc
            with mgr.transaction(scope=scope) as tx:
                for key, value in updates.items():
                    tx.set(key, value)
        else:
            for key, value in updates.items():
                mgr.set(key, value, scope=scope)

    def validate_value(self, provider_id: str, key: str, value: object) -> None:
        """Validate *value* for *key* without persisting it.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        UnknownFieldError
            If *key* is not defined.
        ValidationError
            If *value* is invalid.
        """
        mgr = self._manager(provider_id)
        field = mgr._field_for(key)  # pylint: disable=protected-access
        adapter = TYPE_REGISTRY[field.type].adapter
        try:
            adapter.validate(value, field)
        except (TypeError, ValueError) as exc:
            raise ValidationError(str(exc)) from exc

    def validate_all(self, provider_id: str) -> dict[str, str | None]:
        """Validate all stored values for *provider_id*.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        SigilLoadError
            If configuration values cannot be read.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        raw_map, _ = self.config_backend.read_merged(pid)
        result: dict[str, str | None] = {}
        for field in spec.fields:
            adapter = TYPE_REGISTRY[field.type].adapter
            raw = raw_map.get(field.key)
            try:
                adapter.parse(raw)
            except Exception as exc:  # pragma: no cover - defensive
                result[field.key] = str(exc)
            else:
                result[field.key] = None
        return result

    # ---- Spec persistence ----
    def export_spec(self, provider_id: str, dest: str | Path) -> Path:
        """Export the provider specification to *dest*.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        """
        pid = normalize_provider_id(provider_id)
        spec = self.spec_backend.get_spec(pid)
        path = Path(dest)
        save_provider_spec(path, spec)
        return path

    def reload_spec(self, provider_id: str) -> ProviderSpec:
        """Reload and return the specification for *provider_id*.

        Raises
        ------
        UnknownProviderError
            If the provider is not registered.
        """
        pid = normalize_provider_id(provider_id)
        return self.spec_backend.get_spec(pid)


__all__ = [
    "Orchestrator",
    # errors
    "UnknownProviderError",
    "UnknownFieldError",
    "DuplicateFieldError",
    "ValidationError",
    "PolicyError",
    "SigilLoadError",
]




from pathlib import Path

import pytest

from pysigil.root import (
    Candidate,
    ProjectRootNotFoundError,
    ProjectRootNotFoundWithSuggestionsError,
    find_project_root,
    suggest_candidates,
)

# ---------- helpers ----------

def mkfile(base: Path, rel: str, content: str = "x") -> Path:
    p = base / rel
    p.parent.mkdir(parents=True, exist_ok=True)
    p.write_text(content, encoding="utf-8")
    return p

def mkdir(base: Path, rel: str) -> Path:
    p = base / rel
    p.mkdir(parents=True, exist_ok=True)
    return p

# ---------- core: Tier 0/1/2/3 detection ----------

def test_sentinel_file_detected(tmp_path: Path):
    # /proj/app/.sigil-root (wins immediately at that dir)
    app = tmp_path / "proj" / "app"
    mkdir(app, ".")
    mkfile(app, ".sigil-root", "pinned")
    # Even if a parent has other markers, nearest sentinel in walk wins at that level
    root = find_project_root(start=app / "sub" / "dir")
    assert root == app

def test_tier1_pyproject_detected(tmp_path: Path):
    pkg = tmp_path / "pkg"
    mkfile(pkg, "pyproject.toml", "[project]\nname='x'")
    # Start deeper; should return the dir with pyproject
    start = pkg / "src" / "x" / "deeper"
    mkdir(start, ".")
    root = find_project_root(start=start)
    assert root == pkg

def test_tier1_setup_cfg_detected(tmp_path: Path):
    d = tmp_path / "a" / "b" / "c"
    mkfile(tmp_path / "a" / "b", "setup.cfg", "")
    mkdir(d, ".")
    assert find_project_root(start=d) == tmp_path / "a" / "b"

def test_tier1_git_dir_detected(tmp_path: Path):
    repo = tmp_path / "repo"
    mkdir(repo, ".git")  # directory marker
    start = repo / "x" / "y"
    mkdir(start, ".")
    assert find_project_root(start=start) == repo

def test_tier1_git_file_worktree_detected(tmp_path: Path):
    repo = tmp_path / "repo"
    mkfile(repo, ".git", "gitdir: /worktrees/main")
    mkdir(repo / "src" / "pkg", ".")
    assert find_project_root(start=repo / "src" / "pkg") == repo

def test_nearest_within_tier_wins(tmp_path: Path):
    # pyproject at /proj and at /proj/app; starting at /proj/app/sub should pick /proj/app
    proj = tmp_path / "proj"
    app = proj / "app"
    mkfile(proj, "pyproject.toml", "")
    mkfile(app, "pyproject.toml", "")
    start = app / "sub"
    mkdir(start, ".")
    assert find_project_root(start=start) == app

def test_tier2_tox_when_no_tier1(tmp_path: Path):
    tool = tmp_path / "tooling"
    mkfile(tool, "tox.ini", "")
    start = tool / "src" / "pkg"
    mkdir(start, ".")
    assert find_project_root(start=start) == tool

@pytest.mark.parametrize(
    "fname",
    [
        "WORKSPACE",
        "WORKSPACE.bazel",
        "MODULE.bazel",
        "CMakeLists.txt",
        "package.json",
        "pnpm-workspace.yaml",
        "lerna.json",
        "turbo.json",
        "nx.json",
        "go.mod",
        "Cargo.toml",
        "pom.xml",
        "settings.gradle",
        "settings.gradle.kts",
    ],
)
def test_tier3_other_ecosystems_detected(tmp_path: Path, fname: str):
    root = tmp_path / "mono" / "svc"
    mkfile(root, fname, "")
    start = root / "sub" / "deep"
    mkdir(start, ".")
    assert find_project_root(start=start) == root

def test_tier_precedence_python_over_repo_root(tmp_path: Path):
    # monorepo root has .git; subdir has pyproject → Python (Tier1) should win from within subdir
    mono = tmp_path / "mono"
    mkdir(mono, ".git")
    py = mono / "services" / "py-svc"
    mkfile(py, "pyproject.toml", "")
    start = py / "src" / "x"
    mkdir(start, ".")
    assert find_project_root(start=start) == py

# ---------- IDE (guarded) behavior ----------

def test_ide_guard_accepts_vscode_core_files(tmp_path: Path):
    proj = tmp_path / "quick"
    mkfile(proj, ".vscode/settings.json", "{}")
    # No stronger markers; guarded IDE should pass
    start = proj / "sub"
    mkdir(start, ".")
    assert find_project_root(start=start) == proj

def test_ide_guard_accepts_idea_core_files(tmp_path: Path):
    proj = tmp_path / "quick2"
    mkfile(proj, ".idea/workspace.xml", "<project/>")
    start = proj / "x"
    mkdir(start, ".")
    assert find_project_root(start=start) == proj

def test_ide_guard_accepts_with_co_signals(tmp_path: Path):
    proj = tmp_path / "quick3"
    mkdir(proj, ".vscode")  # empty IDE dir
    mkfile(proj, "README.md", "# hi")  # co-signal
    start = proj / "z"
    mkdir(start, ".")
    assert find_project_root(start=start) == proj

def test_ide_guard_rejects_empty_ide_dir(tmp_path: Path):
    proj = tmp_path / "maybe"
    mkdir(proj, ".vscode")  # empty, no co-signal
    start = proj / "k"
    mkdir(start, ".")
    # Should not auto-select; should raise with suggestions
    with pytest.raises(ProjectRootNotFoundWithSuggestionsError) as e:
        find_project_root(start=start)
    # suggestions should include our dir because .vscode is a weak hint
    cands = e.value.candidates
    assert any(c.path == proj for c in cands)

def test_stronger_marker_trumps_ide(tmp_path: Path):
    proj = tmp_path / "both"
    mkfile(proj, ".vscode/settings.json", "{}")
    mkfile(proj, "pyproject.toml", "")
    # Tier1 should win over IDE
    assert find_project_root(start=proj / "deep") == proj

# ---------- failure / strict / suggestions ----------

def test_strict_false_returns_start(tmp_path: Path):
    nowhere = tmp_path / "nowhere" / "deep"
    mkdir(nowhere, ".")
    assert find_project_root(start=nowhere, strict=False) == nowhere.resolve()

def test_no_markers_raises_with_suggestions(tmp_path: Path):
    # But place some weak hints upward so suggestions are non-empty
    leaf = tmp_path / "a" / "b" / "c"
    mkdir(leaf, ".")
    mkfile(tmp_path / "a", "README.md", "# readme")
    mkfile(tmp_path, "Dockerfile", "FROM scratch")
    with pytest.raises(ProjectRootNotFoundWithSuggestionsError) as e:
        find_project_root(start=leaf)
    cands = e.value.candidates
    assert isinstance(cands, list) and len(cands) > 0
    # top candidate should be either tmp_path (Dockerfile+README at ancestor) or a/ (README)
    assert cands[0].path in {tmp_path, tmp_path / "a"}

# ---------- suggestions ranking & limits ----------

def test_suggestions_rank_by_score_then_proximity(tmp_path: Path):
    # Build a chain: leaf (README), parent (README+Makefile+Dockerfile), top (README)
    leaf = tmp_path / "x" / "y" / "z"
    mkdir(leaf, ".")
    mkfile(leaf, "README.md", "# leaf")
    mid = tmp_path / "x" / "y"
    mkfile(mid, "README.md", "# mid")
    mkfile(mid, "Makefile", "")
    mkfile(mid, "Dockerfile", "")
    top = tmp_path / "x"
    mkfile(top, "README", "top")
    # Mid should score highest and appear first
    cands = suggest_candidates(start=leaf, max_up=5, top_k=5)
    assert cands[0].path == mid
    assert "Dockerfile" in cands[0].hits and "Makefile" in cands[0].hits

def test_suggestions_respect_top_k_and_max_up(tmp_path: Path):
    base = tmp_path / "deep" / "a" / "b" / "c" / "d"
    mkdir(base, ".")
    # sprinkle weak hints across many ancestors
    mkfile(tmp_path / "deep", "README.md", "")
    mkfile(tmp_path / "deep" / "a", "Dockerfile", "")
    mkfile(tmp_path / "deep" / "a" / "b", "Makefile", "")
    mkfile(tmp_path / "deep" / "a" / "b" / "c", ".gitignore", "")
    cands = suggest_candidates(start=base, max_up=3, top_k=2)
    assert len(cands) == 2  # limited by top_k
    # max_up=3 means we only consider c, b, a (not the top 'deep')
    considered_paths = {c.path for c in cands}
    assert tmp_path / "deep" not in considered_paths

# ---------- edge/nuance tests ----------

def test_tier1_vs_tier2_precedence_same_dir(tmp_path: Path):
    d = tmp_path / "mix"
    mkfile(d, "tox.ini", "")
    mkfile(d, "pyproject.toml", "")
    # Same dir: pyproject (Tier1) should win over tox (Tier2), but since nearest dir is the same,
    # the chosen root is d regardless — this just ensures no crash and immediate return.
    assert find_project_root(start=d / "sub") == d

def test_tier3_github_dir_counts_as_root(tmp_path: Path):
    d = tmp_path / "infra"
    mkdir(d, ".github")
    assert find_project_root(start=d / "sub") == d

def test_candidate_model_fields(tmp_path: Path):
    # ensure dataclass is used and hits non-empty when a weak marker exists
    d = tmp_path / "w"
    mkfile(d, "README.md", "")
    cands = suggest_candidates(start=d)
    assert isinstance(cands[0], Candidate)
    assert cands[0].hits and isinstance(cands[0].score, int)

def test_error_types_and_message(tmp_path: Path):
    leaf = tmp_path / "leaf"
    mkdir(leaf, ".")
    with pytest.raises(ProjectRootNotFoundWithSuggestionsError) as e:
        find_project_root(start=leaf)
    err = e.value
    assert isinstance(err, ProjectRootNotFoundError)
    assert str(leaf) in str(err)



from __future__ import annotations

from dataclasses import dataclass

import pytest

from pysigil.settings_metadata import FieldType, TYPE_REGISTRY
from pysigil.ui.options_form import OptionsForm

try:
    import tkinter as tk
    from tkinter import ttk
except Exception:  # pragma: no cover - environment dependent
    tk = None  # type: ignore
    ttk = None  # type: ignore


@dataclass
class DemoOptions:
    text: str
    count: int
    ratio: float
    flag: bool
    tags: list[str]


@dataclass
class IntOptOptional:
    minimum: int | None = None


def _make_form(root):
    ft = FieldType(TYPE_REGISTRY["string"].adapter, option_model=DemoOptions)
    return OptionsForm(root, ft)


@pytest.mark.skipif(tk is None, reason="tkinter not available")
def test_basic_get_set_validate():
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    form = _make_form(root)
    form.set_values(DemoOptions("hello", 3, 0.5, True, ["a", "b"]))
    assert form.get_values() == {
        "text": "hello",
        "count": 3,
        "ratio": 0.5,
        "flag": True,
        "tags": ["a", "b"],
    }
    errs = form.validate()
    assert all(v is None for v in errs.values())
    root.destroy()


@pytest.mark.skipif(tk is None, reason="tkinter not available")
def test_validation_errors():
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    form = _make_form(root)
    form.set_values({"count": "bad"})
    errs = form.validate()
    assert errs["count"]
    root.destroy()


class DummyWidget(ttk.Frame):
    def __init__(self, master):
        super().__init__(master)
        self.entry = ttk.Entry(self)
        self.entry.pack()

    def get_value(self):
        v = self.entry.get()
        return {"value": v} if v else {}

    def set_value(self, value):
        self.entry.delete(0, tk.END)
        if value and "value" in value:
            self.entry.insert(0, value["value"])

    def set_error(self, msg):  # pragma: no cover - placeholder
        pass


def _custom_widget(master):
    return DummyWidget(master)


@pytest.mark.skipif(tk is None, reason="tkinter not available")
def test_custom_option_widget():
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    ft = FieldType(TYPE_REGISTRY["string"].adapter, option_widget=_custom_widget)
    form = OptionsForm(root, ft)
    form.set_values({"value": "foo"})
    assert form.get_values()["value"] == "foo"
    root.destroy()


@pytest.mark.skipif(tk is None, reason="tkinter not available")
def test_optional_field_type():
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    ft = FieldType(TYPE_REGISTRY["integer"].adapter, option_model=IntOptOptional)
    form = OptionsForm(root, ft)
    form.set_values(IntOptOptional(7))
    assert form.get_values()["minimum"] == 7
    form.set_values(IntOptOptional())
    assert form.get_values()["minimum"] is None
    root.destroy()



"""Dialog windows used by :mod:`pysigil.ui.tk`."""

from __future__ import annotations

from collections.abc import Callable

try:  # pragma: no cover - importing tkinter is environment dependent
    import tkinter as tk
    from tkinter import messagebox, ttk
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore
    messagebox = None  # type: ignore
    ttk = None  # type: ignore

from ..aurelia_theme import SCOPE_COLORS, get_palette
from ..provider_adapter import ProviderAdapter, ValueInfo
from ..value_parser import parse_field_value
from .widgets import PillButton


_SCOPE_COLORS = {
    "env": SCOPE_COLORS["Env"],
    "user": SCOPE_COLORS["User"],
    "user-local": SCOPE_COLORS["Machine"],
    "project": SCOPE_COLORS["Project"],
    "project-local": SCOPE_COLORS["ProjectMachine"],
    "default": SCOPE_COLORS["Def"],
}


class EditDialog(tk.Toplevel):  # type: ignore[misc]
    """Simple dialog for editing a key across scopes."""

    def __init__(
        self,
        master: tk.Widget,
        adapter: ProviderAdapter,
        key: str,
        *,
        on_edit_save: Callable[[str, str, object], None] | None = None,
        on_edit_remove: Callable[[str, str], None] | None = None,
    ) -> None:
        super().__init__(master)
        info = adapter.field_info(key)
        label = info.label or key
        self.title(f"Edit — {label}")
        self.transient(master)
        self.grab_set()
        self.resizable(False, False)
        self.adapter = adapter
        self.key = key
        self.on_edit_save = on_edit_save
        self.on_edit_remove = on_edit_remove

        palette = get_palette()
        self.configure(bg=palette["bg"])  # type: ignore[call-arg]

        ttk.Label(self, text=label, style="Title.TLabel").pack(
            anchor="w", padx=18, pady=(12, 0)
        )
        ttk.Label(self, text=key, style="Muted.TLabel").pack(
            anchor="w", padx=18, pady=(0, 6)
        )
        body = ttk.Frame(self, padding=12, style="Card.TFrame")
        body.pack(fill="both", expand=True, padx=18, pady=(0, 12))

        self.entries: dict[str, ttk.Entry] = {}

        values = adapter.values_for_key(key)
        scopes = list(adapter.scopes())
        if "default" not in scopes:
            scopes.append("default")

        _, eff_src = adapter.effective_for_key(key)

        row = 0
        for scope in scopes:
            if scope == "env" and scope not in values:
                continue

            vinfo: ValueInfo | None = values.get(scope)

            def value_provider(s=scope) -> object:
                v = values.get(s)
                if v and v.value is not None:
                    return v.value
                if s == "default":
                    return self.adapter.default_for_key(self.key)
                return None

            can_write = adapter.can_write(scope)
            if scope == "default":
                can_write = False
            elif scope == "env" or adapter.is_overlay(scope):
                can_write = False

            locked = not can_write
            if locked and scope != "default" and not adapter.is_overlay(scope):
                state = "disabled"
            elif eff_src == scope:
                state = "effective"
            elif vinfo and vinfo.value is not None:
                state = "present"
            else:
                state = "empty"

            short_label = adapter.scope_label(scope, short=True)
            long_label = adapter.scope_label(scope, short=False)
            palette = get_palette()
            color = _SCOPE_COLORS.get(scope, palette["ink_muted"])

            pill = PillButton(
                body,
                text=short_label,
                color=color,
                state=state,  # type: ignore[arg-type]
                value_provider=value_provider,
                clickable=False,
                tooltip_title=long_label,
                tooltip_desc=adapter.scope_description(scope),
                locked=locked,
            )
            pill.grid(row=row, column=0, sticky="w", padx=(0, 8), pady=4)

            entry = ttk.Entry(body)
            entry.grid(row=row, column=1, sticky="ew", pady=4)

            if vinfo and vinfo.value is not None:
                entry.insert(0, str(vinfo.value))

            if scope == "default":
                entry.state(["readonly"])
            elif scope == "env" or adapter.is_overlay(scope):
                entry.state(["readonly"])
            elif not can_write:
                entry.state(["disabled"])

            btn_save = ttk.Button(
                body,
                text="Save",
                command=lambda s=scope: self._save_scope(s),
            )
            btn_save.grid(row=row, column=2, padx=4)
            btn_remove = ttk.Button(
                body,
                text="Remove",
                command=lambda s=scope: self._remove_scope(s),
            )
            btn_remove.grid(row=row, column=3, padx=4)

            if not can_write:
                btn_save.state(["disabled"])
                btn_remove.state(["disabled"])

            if vinfo and vinfo.error:
                err = ttk.Label(
                    body,
                    text=vinfo.error,
                    foreground="#b91c1c",
                    style="Card.TLabel",
                )
                err.grid(row=row + 1, column=1, columnspan=3, sticky="w", pady=(0, 4))
                row += 1

            self.entries[scope] = entry
            row += 1
        if info.description_short:
            ttk.Label(
                body,
                text=info.description_short,
                style="CardMuted.TLabel",
                wraplength=400,
                anchor="w",
                justify="left",
            ).grid(row=row, column=0, columnspan=4, sticky="w", pady=(12, 0))
            row += 1
        if info.description:
            ttk.Label(
                body,
                text=info.description,
                style="CardMuted.TLabel",
                wraplength=400,
                anchor="w",
                justify="left",
            ).grid(row=row, column=0, columnspan=4, sticky="w")
            row += 1

        ttk.Button(body, text="Close", command=self.destroy).grid(
            row=row, column=3, sticky="e"
        )
        body.columnconfigure(1, weight=1)

    # -- callbacks ---------------------------------------------------------
    def _save_scope(self, scope: str) -> None:
        if self.on_edit_save is None:
            return

        raw = self.entries[scope].get()
        type_name = self.adapter.field_info(self.key).type

        try:
            value = parse_field_value(type_name, raw)
        except (TypeError, ValueError):
            if messagebox is not None:
                if type_name == "boolean":
                    msg = "Value must be true/false or 1/0"
                else:
                    msg = f"Value must be a {type_name}"
                messagebox.showerror("Invalid value", msg, parent=self)
            return

        try:
            self.on_edit_save(self.key, scope, value)
        except PermissionError as exc:
            if messagebox is not None:
                messagebox.showinfo("Read-only", str(exc), parent=self)
        except Exception as exc:  # pragma: no cover - defensive
            if messagebox is not None:
                messagebox.showerror("Error", str(exc), parent=self)

    def _remove_scope(self, scope: str) -> None:
        if self.on_edit_remove is None:
            return
        try:
            self.on_edit_remove(self.key, scope)
        except PermissionError as exc:
            if messagebox is not None:
                messagebox.showinfo("Read-only", str(exc), parent=self)
            return
        except Exception as exc:  # pragma: no cover - defensive
            if messagebox is not None:
                messagebox.showerror("Error", str(exc), parent=self)
            return
        entry = self.entries.get(scope)
        if entry is not None:
            entry.delete(0, "end")



from pathlib import Path
import pytest

from pysigil.root import ProjectRootNotFoundError, find_project_root


def test_project_root_missing_raises(tmp_path: Path) -> None:
    with pytest.raises(ProjectRootNotFoundError):
        find_project_root(start=tmp_path)



from __future__ import annotations

import os
from typing import Any, Callable, Dict

try:  # pragma: no cover - tkinter may be missing
    import tkinter as tk
    from tkinter import ttk, messagebox
except Exception:  # pragma: no cover
    tk = None  # type: ignore
    ttk = None  # type: ignore
    messagebox = None  # type: ignore

from ..aurelia_theme import SCOPE_COLORS, get_palette
from ..provider_adapter import ProviderAdapter, ValueInfo
from .widgets import PillButton, HoverTip

# Mapping from scope ids to color constants used by :class:`PillButton`
_SCOPE_COLORS = {
    "env": SCOPE_COLORS["Env"],
    "user": SCOPE_COLORS["User"],
    "user-local": SCOPE_COLORS["Machine"],
    "project": SCOPE_COLORS["Project"],
    "project-local": SCOPE_COLORS["ProjectMachine"],
    "default": SCOPE_COLORS["Def"],
}


def _debug_columns() -> bool:
    """Return True when the debug column overlay is enabled."""
    return bool(os.environ.get("PYSGIL_DEBUG_COLUMNS"))


class FieldRow(tk.Frame):
    """Representation of a single field row with scope pills."""

    def __init__(
        self,
        master: tk.Widget,
        adapter: ProviderAdapter,
        key: str,
        on_pill_click: Callable[[str, str], None],
        *,
        compact: bool = True,
        on_edit_click: Callable[[str], None] | None = None,
    ) -> None:
        palette = get_palette()
        super().__init__(master, bg=palette["card"], highlightthickness=0)
        self.adapter = adapter
        self.key = key
        self._on_pill_click = on_pill_click
        self._on_edit_click = on_edit_click
        self.compact = compact

        ink = palette["ink"]
        ink_muted = palette["ink_muted"]
        field = palette["field"]

        # container for key + info button
        self.key_frame = ttk.Frame(self, style="CardBody.TFrame")

        self.key_frame.grid(row=0, column=0, sticky="w", pady=(6, 0))

        info = None
        if hasattr(adapter, "field_info"):
            try:
                info = adapter.field_info(key)  # type: ignore[attr-defined]
            except Exception:
                info = None
        label = getattr(info, "label", None) or key
        self.lbl_key = ttk.Label(
            self.key_frame,
            text=label,
            padding=6,
            style="CardKey.TLabel",
        )
        self.lbl_key.pack(side="left")

        self.info_btn: tk.Label | None = None
        self.lbl_desc: ttk.Label | None = None
        if info:
            desc = info.description or info.description_short
            tip_lines = [f"Key: {key}"]
            if desc:
                tip_lines.append(desc)
            tip = "\n\n".join(tip_lines)
            if desc or key:
                self.info_btn = tk.Label(
                    self.key_frame,
                    text="\u24D8",
                    fg=ink_muted,
                    cursor="question_arrow",
                    bg=palette["card"],
                    takefocus=1,
                )
                self.info_btn.pack(side="left", padx=(4, 0))
                HoverTip(self.info_btn, lambda: tip)
                HoverTip(self.lbl_key, lambda: tip)
            if info.description_short:
                self.lbl_desc = ttk.Label(
                    self,
                    text=info.description_short,
                    style="CardMuted.TLabel",
                    wraplength=600,
                    padding=(6,0,0,0) # (left, top, right, bottom)
                )
                self.lbl_desc.grid(row=1, column=0, columnspan=3, sticky="w")

        # effective value display
        self.var_eff = tk.StringVar(value="") if tk else None
        self.lbl_eff = tk.Label(
            self,
            textvariable=self.var_eff,
            bg=field,
            fg=ink,
            bd=1,
            relief="ridge",
            padx=10,

            pady=0,
            anchor="nw",
        )
        self.lbl_eff.grid(row=0, column=1, sticky="new", padx=(8, 8), pady=(6, 0))

        # container for scope pills
        self.pills = ttk.Frame(self, style="CardBody.TFrame")
        self.pills.grid(row=0, column=2, sticky="nw", pady=(6, 0))


        # edit action button
        self.btn_edit = ttk.Button(
            self,
            text="Edit…",
            command=lambda: self._on_edit_click(self.key) if self._on_edit_click else None,
        )

        self.btn_edit.grid(row=0, column=3, padx=4, pady=(6, 0), sticky="n")

        # ensure value box matches edit button height
        if tk is not None:
            self.after_idle(self._sync_value_height)


        self.columnconfigure(1, weight=1)

        self._pill_widgets: Dict[str, PillButton] = {}

        self.refresh()

        if _debug_columns() and tk is not None:
            self._debug_canvas = tk.Canvas(self, highlightthickness=0)
            self._debug_canvas.place(relx=0, rely=0, relwidth=1, relheight=1)
            self.bind("<Configure>", self._on_debug_configure)

    # ------------------------------------------------------------------
    def set_compact(self, compact: bool) -> None:
        """Toggle compact mode and rebuild pills."""
        if self.compact != compact:
            self.compact = compact
            self.refresh()

    # ------------------------------------------------------------------
    def refresh(self) -> None:
        """Refresh the effective value and update scope pills."""
        if tk is None:  # pragma: no cover - defensive
            return

        # effective value -------------------------------------------------------
        eff_val, eff_src = self.adapter.effective_for_key(self.key)
        val_txt = "—" if eff_val is None else str(eff_val)
        if eff_src is None:
            src_txt = "—"
        else:
            src_txt = self.adapter.scope_label(eff_src)
        self.var_eff.set(f"{val_txt}  ({src_txt})")

        values: Dict[str, ValueInfo] = self.adapter.values_for_key(self.key)
        scopes = self.adapter.scopes()

        # Ensure pills are packed in the same order as ``scopes`` by first
        # unpacking any existing widgets.  Without this, newly created scopes
        # would always be appended to the end regardless of the desired order.
        for child in self.pills.pack_slaves():
            child.pack_forget()

        for scope in scopes:
            has_value = scope in values
            def value_provider(s=scope) -> Any:
                if s in values:
                    return values[s].value
                if s == "default":
                    return self.adapter.default_for_key(self.key)
                return None
            can_write = self.adapter.can_write(scope)
            self.update_pill(
                scope,
                effective=(eff_src == scope),
                present=has_value,
                can_write=can_write,
                value_provider=value_provider,
            )

        # adjust value label height when geometry might change
        self._sync_value_height()

    def _sync_value_height(self) -> None:
        if tk is None:  # pragma: no cover - defensive
            return
        try:
            # ``winfo_height`` reflects the actual rendered size whereas
            # ``winfo_reqheight`` only reports the requested size.  Using the
            # real height keeps the effective value box in sync with themed
            # button and pill widgets whose final size may exceed their
            # request due to style padding.
            self.update_idletasks()
            btn_h = self.btn_edit.winfo_height()
            lbl_h = self.lbl_eff.winfo_height()
            diff = btn_h - lbl_h
            if diff > 0:
                pad = (diff+1) // 5
                self.lbl_eff.configure(pady=pad)
        except Exception:
            pass

    def _on_debug_configure(self, event: tk.Event) -> None:
        if not _debug_columns():
            return
        canvas = getattr(self, "_debug_canvas", None)
        if canvas is None:
            return
        canvas.delete("all")
        cols = self.grid_size()[0]
        for col in range(cols - 1):
            try:
                x, y, w, h = self.grid_bbox(col, 0)
            except Exception:
                continue
            canvas.create_line(x + w, 0, x + w, event.height, fill="red", width=2)

    def update_pill(
        self,
        name: str,
        *,
        effective: bool,
        present: bool,
        can_write: bool,
        value_provider: Callable[[], Any],
    ) -> None:
        """Update or create a single pill widget.

        ``effective`` indicates whether the pill represents the effective
        scope while ``present`` reports if the scope has an explicit value.
        ``can_write`` controls the clickable state and ``value_provider`` is
        used for tooltip display.
        """
        if tk is None:  # pragma: no cover - defensive
            return

        if (
            name == "default" and not present and self.compact
        ) or (self.compact and name != "default" and not present):
            pill = self._pill_widgets.get(name)
            if pill and pill.winfo_ismapped():
                pill.pack_forget()
            return

        locked = not can_write
        if (not can_write and name != "default" and not self.adapter.is_overlay(name)):
            state = "disabled"
        elif effective:
            state = "effective"
        elif present:
            state = "present"
        else:
            state = "empty"

        short_label = self.adapter.scope_label(name, short=True)
        long_label = self.adapter.scope_label(name, short=False)
        palette = get_palette()
        color = _SCOPE_COLORS.get(name, palette["ink_muted"])

        def cb() -> None:
            if not locked and self._on_pill_click:
                self._on_pill_click(self.key, name)
            elif locked:
                hint = self.adapter.scope_hint(name)
                if messagebox is not None and hint:
                    try:
                        messagebox.showinfo("Read-only", hint)
                    except Exception:
                        pass

        pill = self._pill_widgets.get(name)
        if pill is None:
            pill = PillButton(
                self.pills,
                text=short_label,
                color=color,
                state=state,  # type: ignore[arg-type]
                value_provider=value_provider,
                clickable=True,
                on_click=cb,
                tooltip_title=long_label,
                tooltip_desc=self.adapter.scope_description(name),
                locked=locked,
            )
            self._pill_widgets[name] = pill
        else:
            pill.text = short_label
            pill.color = color
            pill.state = state  # type: ignore[assignment]
            pill.locked = locked
            pill.clickable = True
            pill.value_provider = value_provider
            pill.tooltip_title = long_label
            pill.tooltip_desc = self.adapter.scope_description(name)
            pill.on_click = cb
            pill.bind("<Button-1>", lambda e: cb())
            pill.configure(cursor="hand2")
            pill._draw()

        if not pill.winfo_ismapped():
            pill.pack(side="left", padx=(0, 6))

    # ------------------------------------------------------------------
    def update_metadata(self, info: object) -> None:  # pragma: no cover - simple
        """Update field key label based on new metadata."""
        key = getattr(info, "key", self.key)
        self.key = key
        label = getattr(info, "label", None) or key
        self.lbl_key.configure(text=label)
        desc = getattr(info, "description", None) or getattr(info, "description_short", None)
        tip_lines = [f"Key: {key}"]
        if desc:
            tip_lines.append(desc)
        tip = "\n\n".join(tip_lines)
        if self.info_btn is not None:
            HoverTip(self.info_btn, lambda: tip)
        HoverTip(self.lbl_key, lambda: tip)


__all__ = ["FieldRow"]



import importlib
import pytest
import types

try:
    import tkinter as tk
except Exception:  # pragma: no cover - tkinter missing
    tk = None  # type: ignore

import pysigil.ui.tk.rows as tk_rows
from pysigil.ui.tk.rows import FieldRow
from pysigil.ui.tk.widgets import PillButton
from pysigil.ui.provider_adapter import ValueInfo


class DummyAdapter:
    def scopes(self):
        return ["env", "user", "user-local", "project", "project-local", "default"]

    def scope_label(self, scope_id, short=False):
        short_labels = {
            "env": "Env",
            "user": "User",
            "user-local": "Machine",
            "project": "Project",
            "project-local": "Project·Machine",
            "default": "Default",
        }
        long_labels = {
            "env": "Environment",
            "user": "User",
            "user-local": "Machine",
            "project": "Project",
            "project-local": "Project on this Machine",
            "default": "Default",
        }
        return (short_labels if short else long_labels)[scope_id]

    _DESCRIPTIONS = {
        "env": "Environment override",
        "user": "User configuration",
        "user-local": "Machine specific user setting",
        "project": "Shared project configuration",
        "project-local": "Project setting for this machine",
        "default": "Built-in default",
    }

    def scope_description(self, scope_id):
        return self._DESCRIPTIONS[scope_id]

    def can_write(self, scope_id):
        return scope_id not in {"env", "project", "default"}

    def is_overlay(self, scope_id):
        return scope_id == "env"

    def values_for_key(self, key):
        return {
            "env": ValueInfo("e"),
            "user": ValueInfo("u"),
            "project-local": ValueInfo("pl"),
            "default": ValueInfo("d"),
        }

    def effective_for_key(self, key):
        return "u", "user"

    def default_for_key(self, key):
        return "d"

    def scope_hint(self, scope_id):
        if scope_id == "default":
            return "Default scope is read-only. Use Author Tools to change it."
        label = self.scope_label(scope_id)
        return f"{label} scope is read-only. Use Author Tools to change it."


def _collect_pills(row):
    return [w for w in row.pills.winfo_children() if isinstance(w, PillButton)]


def test_field_row_full_mode(monkeypatch):
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    clicks = []
    adapter = DummyAdapter()
    row = FieldRow(root, adapter, "alpha", lambda k, s: clicks.append((k, s)), compact=False)
    pills = _collect_pills(row)
    states = {p.text: p.state for p in pills}
    assert states == {
        "Env": "present",
        "User": "effective",
        "Machine": "empty",
        "Project": "disabled",
        "Project·Machine": "present",
        "Default": "present",
    }
    default_pill = next(p for p in pills if p.text == "Default")
    assert default_pill.locked
    env_pill = next(p for p in pills if p.text == "Env")
    assert env_pill.locked
    user_pill = next(p for p in pills if p.text == "User")
    assert adapter.scope_description("user") in user_pill._tip_text()
    for p in _collect_pills(row):
        if p.text == "User":
            p.on_click()
            break
    assert clicks == [("alpha", "user")]
    lock_txt = row.var_eff.get()
    assert lock_txt.startswith("u")
    assert "(User)" in lock_txt
    root.destroy()


def test_field_row_compact_mode():
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    row = FieldRow(root, DummyAdapter(), "alpha", lambda k, s: None, compact=True)
    pills = _collect_pills(row)
    assert [p.text for p in pills] == ["Env", "User", "Project·Machine", "Default"]
    root.destroy()


def test_locked_pill_shows_hint(monkeypatch):
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    calls = []
    dummy_box = types.SimpleNamespace(showinfo=lambda *a, **k: calls.append(a))
    monkeypatch.setattr(tk_rows, "messagebox", dummy_box)
    row = FieldRow(root, DummyAdapter(), "alpha", lambda k, s: None, compact=False)
    env_pill = next(p for p in _collect_pills(row) if p.text == "Env")
    env_pill.on_click()
    assert calls and "Author" in calls[0][1]
    root.destroy()


class NoDefaultAdapter(DummyAdapter):
    def values_for_key(self, key):
        return {"env": ValueInfo("e")}

    def effective_for_key(self, key):
        return "e", "env"

    def default_for_key(self, key):
        return None


def test_field_row_hides_default_when_missing():
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    row = FieldRow(root, NoDefaultAdapter(), "alpha", lambda k, s: None, compact=True)
    pills = [p.text for p in _collect_pills(row)]
    assert "Default" not in pills
    root.destroy()


def test_field_row_shows_default_when_missing_in_full_mode():
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    row = FieldRow(root, NoDefaultAdapter(), "alpha", lambda k, s: None, compact=False)
    pills = {p.text: p.state for p in _collect_pills(row)}
    assert pills["Default"] == "empty"
    root.destroy()



class DefaultOnlyAdapter(DummyAdapter):
    def scopes(self):
        return ["default"]

    def values_for_key(self, key):
        return {"default": ValueInfo("d")}

    def effective_for_key(self, key):
        return "d", "default"


def test_field_row_default_effective():
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")
    row = FieldRow(root, DefaultOnlyAdapter(), "alpha", lambda k, s: None, compact=True)
    pills = _collect_pills(row)
    assert len(pills) == 1
    pill = pills[0]
    assert pill.text == "Default"
    assert pill.state == "effective"
    assert pill.color == "#000000"
    assert pill.locked
    root.destroy()


def test_debug_columns_env_var_after_import(monkeypatch):
    if tk is None:
        pytest.skip("tkinter not available")
    try:
        root = tk.Tk()
    except Exception:
        pytest.skip("no display available")

    monkeypatch.delenv("PYSGIL_DEBUG_COLUMNS", raising=False)
    importlib.reload(tk_rows)

    monkeypatch.setenv("PYSGIL_DEBUG_COLUMNS", "1")
    row = tk_rows.FieldRow(root, DummyAdapter(), "alpha", lambda k, s: None, compact=False)
    assert hasattr(row, "_debug_canvas")
    root.destroy()




from __future__ import annotations

import os
import sys
import tkinter as tk
import webbrowser
from pathlib import Path
from tkinter import filedialog, messagebox, ttk

from ...authoring import (
    DevLinkError,
    _dev_dir,
    ensure_sigil_package_data,
    link,
    normalize_provider_id,
)
from ...resolver import (
    ensure_defaults_file,
    find_package_dir,
    read_dist_name_from_pyproject,
    validate_package_dir,
)
from ...root import ProjectRootNotFoundError, find_project_root
from ..aurelia_theme import get_palette, use

APP_TITLE = "Sigil – Register Package Defaults"




class RegisterApp(tk.Tk):
    def __init__(self) -> None:
        super().__init__()
        use(self)
        palette = get_palette()

        style = ttk.Style(self)
        style.configure("TLabel",font=(None, 10, "bold"))

        self.configure(bg=palette["bg"])  # type: ignore[call-arg]
        self.title(APP_TITLE)
        self.geometry("640x320")
        self.minsize(520, 280)
        self.option_add("*highlightcolor", palette["gold"])
        self.option_add("*highlightbackground", palette["bg"])
        self.option_add("*highlightthickness", 1)

        self.defaults_path: tk.StringVar = tk.StringVar(value="")
        self.provider_id: tk.StringVar = tk.StringVar(value="")
        self.message_var: tk.StringVar = tk.StringVar(
            value="Pick your package folder (contains __init__.py). Provider ID is pre-filled; edit if needed."
        )

        self._build()
        self._try_autodetect()

    def _build(self) -> None:
        pad = {"padx": 12, "pady": 8}

        frm = ttk.Frame(self)
        frm.pack(fill=tk.BOTH, expand=True)

        row1 = ttk.Frame(frm)
        row1.pack(fill=tk.X, **pad)
        ttk.Label(row1, text="1) Package folder (contains __init__.py):").pack(anchor=tk.W)

        pick = ttk.Frame(row1)
        pick.pack(fill=tk.X)
        entry = ttk.Entry(pick, textvariable=self.defaults_path)
        entry.pack(side=tk.LEFT, fill=tk.X, expand=True)
        ttk.Button(pick, text="Browse…", command=self.on_browse).pack(side=tk.LEFT, padx=6)

        row2 = ttk.Frame(frm)
        row2.pack(fill=tk.X, **pad)
        ttk.Label(row2, text="2) Provider ID (PEP 503 normalized):").pack(anchor=tk.W)
        prov = ttk.Entry(row2, textvariable=self.provider_id)
        prov.pack(fill=tk.X)

        actions = ttk.Frame(frm)
        actions.pack(fill=tk.X, **pad)
        ttk.Button(actions, text="Register", command=self.on_register).pack(side=tk.LEFT)
        ttk.Button(actions, text="Open dev-links folder", command=self.on_open_folder).pack(
            side=tk.LEFT, padx=8
        )
        ttk.Button(actions, text="Quit", command=self.destroy).pack(side=tk.RIGHT)

        status = ttk.Frame(self)
        status.pack(fill=tk.X)
        ttk.Label(status, textvariable=self.message_var).pack(anchor=tk.W, padx=10, pady=6)

    def on_browse(self) -> None:
        folder = filedialog.askdirectory(
            title="Select your package folder (e.g., mypkg/ or src/mypkg/)",
            mustexist=True,
            initialdir=os.getcwd(),
        )
        if not folder:
            return

        chosen = Path(folder)
        try:
            pkg = validate_package_dir(chosen)
        except ValueError as exc:
            messagebox.showerror(
                "Pick your package folder",
                f"{exc}\n\nPick the folder that contains __init__.py.",
            )
            return

        self.defaults_path.set(str(pkg))
        import_name = pkg.name
        if not self.provider_id.get().strip():
            self.provider_id.set(normalize_provider_id(import_name))
        self.message_var.set(f"Package folder: {pkg}")

    def _try_autodetect(self) -> None:
        cwd = Path(os.getcwd())
        try:
            root = find_project_root(cwd)
        except ProjectRootNotFoundError:
            root = cwd
        dist_name = read_dist_name_from_pyproject(root)
        pkg = find_package_dir(root, dist_name)
        if pkg:
            try:
                pkg = validate_package_dir(pkg)
            except ValueError:
                return
            self.defaults_path.set(str(pkg))
            if not self.provider_id.get().strip():
                self.provider_id.set(normalize_provider_id(pkg.name))
            self.message_var.set(f"Detected package folder: {pkg}")

    def on_register(self) -> None:
        pkg_dir_raw = self.defaults_path.get().strip()
        provider = self.provider_id.get().strip()

        if not pkg_dir_raw:
            messagebox.showerror("Missing folder", "Please choose your **package folder**.")
            return
        if not provider:
            messagebox.showerror("Missing provider id", "Please confirm or edit the provider id.")
            return

        try:
            pkg = validate_package_dir(Path(pkg_dir_raw))
        except ValueError as exc:
            messagebox.showerror(
                "Invalid folder",
                f"{exc}\n\nPick the folder that contains __init__.py.",
            )
            return

        import_name = pkg.name
        if not provider:
            provider = normalize_provider_id(import_name)
            self.provider_id.set(provider)

        provider_norm = normalize_provider_id(provider)
        try:
            settings_path = ensure_defaults_file(pkg, provider_norm)
            dl = link(provider_norm, settings_path, validate=True)
            self.message_var.set(f"Linked {dl.provider_id} -> {dl.defaults_path}")
            try:
                project_root = find_project_root(pkg)
            except ProjectRootNotFoundError:
                project_root = pkg
            ensure_sigil_package_data(project_root, import_name)
            messagebox.showinfo(
                "Success",
                f"Created/verified {settings_path}\n\nRegistered dev link for {dl.provider_id}.",
            )
        except DevLinkError as e:
            self.message_var.set(str(e))
            messagebox.showerror("Error", str(e))

    def on_open_folder(self) -> None:
        folder = _dev_dir()
        folder.mkdir(parents=True, exist_ok=True)
        if sys.platform.startswith("win"):
            os.startfile(folder)  # type: ignore[attr-defined]
        elif sys.platform == "darwin":
            os.system(f"open '{folder}'")
        else:
            try:
                os.system(f"xdg-open '{folder}'")
            except Exception:
                webbrowser.open(str(folder))


def main() -> None:
    app = RegisterApp()
    app.mainloop()


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path

import pytest

from pysigil.resolver import (
    default_provider_id,
    ensure_defaults_file,
    find_package_dir,
    validate_package_dir,
    read_dist_name_from_pyproject,
)
from pysigil.root import find_project_root


def write_pyproject(root: Path, name: str | None = None) -> None:
    content = "[project]\n"
    if name is not None:
        content += f"name = \"{name}\"\n"
    (root / "pyproject.toml").write_text(content, encoding="utf-8")


def test_find_project_root_pyproject(tmp_path: Path) -> None:
    write_pyproject(tmp_path)
    sub = tmp_path / "sub"
    sub.mkdir()
    assert find_project_root(sub) == tmp_path


def test_provider_from_pyproject(tmp_path: Path) -> None:
    write_pyproject(tmp_path, name="My_Package.Name")
    name = read_dist_name_from_pyproject(tmp_path)
    assert name == "My_Package.Name"
    pkg = tmp_path / "my_pkg"
    pkg.mkdir()
    (pkg / "__init__.py").touch()
    assert (
        default_provider_id(pkg, name) == "my-pkg"
    )


def test_validate_package_dir(tmp_path: Path) -> None:
    pkg = tmp_path / "pkg"
    pkg.mkdir()
    (pkg / "__init__.py").touch()
    assert validate_package_dir(pkg) == pkg.resolve()
    with pytest.raises(ValueError):
        validate_package_dir(tmp_path / "missing")
    other = tmp_path / "no_init"
    other.mkdir()
    with pytest.raises(ValueError):
        validate_package_dir(other)


def test_find_package_src_layout(tmp_path: Path) -> None:
    write_pyproject(tmp_path, name="my-pkg")
    pkg = tmp_path / "src" / "my_pkg"
    pkg.mkdir(parents=True)
    (pkg / "__init__.py").touch()
    found = find_package_dir(tmp_path, "my-pkg")
    assert found == pkg


def test_find_package_flat_layout(tmp_path: Path) -> None:
    write_pyproject(tmp_path, name="my-pkg")
    pkg = tmp_path / "my_pkg"
    pkg.mkdir()
    (pkg / "__init__.py").touch()
    found = find_package_dir(tmp_path, "my-pkg")
    assert found == pkg


def test_single_package_under_src(tmp_path: Path) -> None:
    write_pyproject(tmp_path)
    pkg = tmp_path / "src" / "only_pkg"
    pkg.mkdir(parents=True)
    (pkg / "__init__.py").touch()
    found = find_package_dir(tmp_path, None)
    assert found == pkg


def test_ambiguous_packages(tmp_path: Path) -> None:
    write_pyproject(tmp_path)
    for name in ["a", "b"]:
        pkg = tmp_path / "src" / name
        pkg.mkdir(parents=True)
        (pkg / "__init__.py").touch()
    assert find_package_dir(tmp_path, None) is None


def test_ensure_defaults_file(tmp_path: Path) -> None:
    pkg = tmp_path / "pkg"
    pkg.mkdir()
    (pkg / "__init__.py").touch()
    ini = ensure_defaults_file(pkg, "prov")
    text = ini.read_text(encoding="utf-8")
    assert "[prov]" in text



"""Aurelia ttk theme palette definitions."""

from __future__ import annotations

from collections.abc import Mapping

from .theme import (
    ThemeSpec,
    apply_theme,
    get_active_palette,
    register_scope_styles as _register_scope_styles,
)

try:  # pragma: no cover - importing tkinter is environment dependent
    import tkinter as tk
    from tkinter import ttk
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore
    ttk = None  # type: ignore

_AURELIA_PALETTE: dict[str, object] = {
    "bg": "#2B313B",
    "card": "#EEF3FA",
    "card_edge": "#D4DEEE",
    "hdr_fg": "#F4F8FF",
    "hdr_muted": "#C3CFE4",
    "ink": "#0E1724",
    "ink_muted": "#586A84",
    "field": "#FFFFFF",
    "field_bd": "#C8D3E6",
    "gold": "#D4B76A",
    "gold_hi": "#f1d48e",
    "title_accent": "#f1d48e",
    "primary": "#365DC6",
    "primary_hover": "#587BE2",
    "on_primary": "#F7FAFF",
    "tooltip_bg": "#0E1724",
    "tooltip_fg": "#F7FAFF",
    "scopes": {
        "Env": "#1f9d66",  # green-700
        "User": "#1e40af",  # blue-900
        "Machine": "#1c6e7d",  # emerald-900
        "Project": "#D4B76A",  # violet-700
        "ProjectMachine": "#c25e0c",  # orange-700
        "Def": "#334155",  # slate-700
    },
}

SCOPE_COLORS: dict[str, str] = _AURELIA_PALETTE["scopes"]  # type: ignore[index]

THEME = ThemeSpec(name="aurelia", ttk_theme="clam", palette=_AURELIA_PALETTE)


def get_palette() -> dict[str, object]:
    """Return the palette for the active Aurelia theme."""

    palette = get_active_palette()
    return palette if palette else _AURELIA_PALETTE


def use(root: tk.Misc, *, palette: Mapping[str, object] | None = None) -> None:
    """Apply the Aurelia theme to *root*."""

    apply_theme(root, THEME, palette_override=palette)


def register_scope_styles(style: ttk.Style, scope_colors: Mapping[str, str]) -> None:
    """Register scope-specific styles for Aurelia."""

    _register_scope_styles(style, scope_colors)


__all__ = ["THEME", "use", "get_palette", "register_scope_styles", "SCOPE_COLORS"]



"""Utility widgets for editing lists in tkinter.

This module provides :class:`ListEditor`, a small composite widget used in
`pysigil`'s tkinter UI to edit lists of values.  The widget supports three
modes:

``"simple"``
    A single column list of values (strings by default).
``"kv"``
    Two columns labelled ``key`` and ``value``.
``"choice"``
    Two columns labelled ``value`` and ``label`` suitable for enum choices.

The implementation intentionally keeps the feature set light weight – it is
primarily intended for tests and manual usage.  It nevertheless exposes a
couple of niceties such as duplicate removal and alphabetic sorting.  The
widget emits ``<<ListChanged>>`` whenever its contents change and offers
``get_list`` / ``set_list`` helpers for converting between the visual list and
Python data structures.

The companion :class:`ListEditDialog` wraps the editor in a modal ``Toplevel``
window and exposes an ``result`` attribute containing the edited list when the
user confirms the dialog.
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Literal, Sequence

try:  # pragma: no cover - importing tkinter is environment dependent
    import tkinter as tk
    from tkinter import ttk, simpledialog, messagebox, filedialog
except Exception:  # pragma: no cover - fallback when tkinter missing
    tk = None  # type: ignore
    ttk = None  # type: ignore
    simpledialog = None  # type: ignore
    messagebox = None  # type: ignore
    filedialog = None  # type: ignore

import csv
from io import StringIO

from ..aurelia_theme import get_palette, use


Mode = Literal["simple", "kv", "choice"]


@dataclass
class Column:
    """Specification for a Treeview column."""

    id: str
    heading: str
    width: int = 140


class ListEditor(ttk.Frame):  # pragma: no cover - exercised via tk tests
    """Edit a list of values inside a :mod:`tkinter` application.

    Parameters largely mirror the spec from :mod:`AGENTS.md` and have been
    trimmed to fit within the repository's testing needs.  Normalisation and
    validation hooks can be supplied to influence how values are stored.
    """

    def __init__(
        self,
        master: tk.Widget,
        *,
        mode: Mode = "simple",
        value: Sequence[Any] | None = None,
        unique: bool = True,
        allow_empty: bool = False,
        max_items: int | None = None,
        validator: Callable[[Any], tuple[bool, str | None]] | None = None,
        normalizer: Callable[[Any], Any] | None = None,
        columns: Sequence[Column] | None = None,
        allow_reorder: bool = True,
        allow_sort: bool = True,
        allow_paste: bool = True,
        allow_import_export: bool = True,
    ) -> None:
        if tk is None:  # pragma: no cover - environment guard
            raise RuntimeError("tkinter is required for ListEditor")
        super().__init__(master, style="CardBody.TFrame")
        self.mode = mode
        self.unique = unique
        self.allow_empty = allow_empty
        self.max_items = max_items
        self.validator = validator
        self.normalizer = normalizer
        self.allow_reorder = allow_reorder
        self.allow_sort = allow_sort
        self.allow_paste = allow_paste
        self.allow_import_export = allow_import_export
        self._items: list[Any] = []
        self._build(columns)
        self.set_list(list(value or []))

    # ------------------------------------------------------------------
    # UI construction
    # ------------------------------------------------------------------
    def _build(self, columns: Sequence[Column] | None) -> None:
        toolbar = ttk.Frame(self, style="CardBody.TFrame")
        toolbar.pack(fill="x", padx=2, pady=2)
        ttk.Button(toolbar, text="Add", command=self._on_add).pack(side="left")
        ttk.Button(toolbar, text="Edit", command=self._on_edit).pack(side="left")
        ttk.Button(toolbar, text="Remove", command=self._on_remove).pack(side="left")
        if self.allow_reorder:
            ttk.Button(toolbar, text="↑", command=lambda: self._move(-1)).pack(
                side="left", padx=(6, 0)
            )
            ttk.Button(toolbar, text="↓", command=lambda: self._move(1)).pack(
                side="left"
            )
        if self.allow_sort:
            ttk.Button(toolbar, text="Sort", command=self.sort_items).pack(
                side="left", padx=(6, 0)
            )
        ttk.Button(toolbar, text="Dedupe", command=self.dedupe).pack(side="left")
        if self.allow_paste:
            ttk.Button(toolbar, text="Paste", command=self._on_paste).pack(
                side="left", padx=(6, 0)
            )
        if self.allow_import_export:
            ttk.Button(toolbar, text="Import CSV…", command=self._on_import).pack(
                side="left", padx=(6, 0)
            )
            ttk.Button(toolbar, text="Export CSV…", command=self._on_export).pack(
                side="left"
            )

        # determine columns
        if columns is None:
            if self.mode == "simple":
                columns = [Column("value", "Value", 200)]
            elif self.mode == "kv":
                columns = [Column("key", "Key"), Column("value", "Value")]
            else:  # choice
                columns = [Column("value", "Value"), Column("label", "Label")]
        self._columns = list(columns)

        self._tree = ttk.Treeview(
            self,
            columns=[c.id for c in self._columns],
            show="headings",
            selectmode="extended",
            height=8,
        )
        for col in self._columns:
            self._tree.heading(col.id, text=col.heading)
            self._tree.column(col.id, width=col.width, anchor="w")
        self._tree.pack(fill="both", expand=True, padx=2, pady=2)
        self._tree.bind("<Double-1>", lambda e: self._on_edit())

    # ------------------------------------------------------------------
    # List manipulation helpers
    # ------------------------------------------------------------------
    def get_list(self) -> list[Any]:
        """Return the currently edited list."""

        return list(self._items)

    def set_list(self, lst: Sequence[Any]) -> None:
        """Replace the contents of the editor with ``lst``."""

        self._items = list(lst)
        self._refresh_tree()
        self.event_generate("<<ListChanged>>")

    def set_readonly(self, flag: bool) -> None:
        """Enable/disable editing."""

        state = "disabled" if flag else "!disabled"
        for child in self.winfo_children():
            try:
                child.state([state])  # type: ignore[call-arg]
            except Exception:
                pass

    # -- internals -----------------------------------------------------
    def _norm(self, item: Any) -> Any:
        return self.normalizer(item) if self.normalizer else item

    def _validate(self, item: Any) -> tuple[bool, str | None]:
        if self.validator:
            return self.validator(item)
        if not self.allow_empty and item in ("", None):
            return False, "Empty values are not allowed"
        return True, None

    def _refresh_tree(self) -> None:
        self._tree.delete(*self._tree.get_children())
        for idx, item in enumerate(self._items):
            values = self._item_to_row(item)
            self._tree.insert("", "end", iid=str(idx), values=values)

    def _item_to_row(self, item: Any) -> Sequence[str]:
        if self.mode == "simple":
            return [str(item)]
        elif self.mode == "kv":
            return [str(item.get("key", "")), str(item.get("value", ""))]
        else:  # choice
            return [str(item.get("value", "")), str(item.get("label", ""))]

    # ------------------------------------------------------------------
    # Commands
    # ------------------------------------------------------------------
    def _on_add(self) -> None:
        item = self._prompt_item()
        if item is None:
            return
        if self.max_items is not None and len(self._items) >= self.max_items:
            return
        norm = self._norm(item)
        ok, _ = self._validate(norm)
        if not ok:
            return
        if self.unique and self._is_duplicate(norm):
            return
        self._items.append(norm)
        self._refresh_tree()
        self.event_generate("<<ListChanged>>")

    def _on_edit(self) -> None:
        sel = self._tree.selection()
        if not sel:
            return
        idx = int(sel[0])
        item = self._items[idx]
        new_item = self._prompt_item(item)
        if new_item is None:
            return
        norm = self._norm(new_item)
        ok, _ = self._validate(norm)
        if not ok:
            return
        if self.unique:
            tmp = list(self._items)
            tmp[idx] = norm
            if self._is_duplicate(norm, tmp, idx):
                return
        self._items[idx] = norm
        self._refresh_tree()
        self.event_generate("<<ListChanged>>")

    def _on_remove(self) -> None:
        indices = sorted((int(i) for i in self._tree.selection()), reverse=True)
        for idx in indices:
            del self._items[idx]
        self._refresh_tree()
        if indices:
            self.event_generate("<<ListChanged>>")

    def _move(self, offset: int) -> None:
        if offset not in (-1, 1):
            return
        indices = sorted((int(i) for i in self._tree.selection()))
        if offset > 0:
            indices.reverse()
        moved = False
        new_selection: list[int] = []
        for idx in indices:
            new_idx = idx + offset
            if not (0 <= new_idx < len(self._items)):
                new_selection.append(idx)
                continue
            self._items[idx], self._items[new_idx] = (
                self._items[new_idx],
                self._items[idx],
            )
            new_selection.append(new_idx)
            moved = True
        if moved:
            self._refresh_tree()
            self._tree.selection_set([str(i) for i in sorted(new_selection)])
            self.event_generate("<<ListChanged>>")

    def sort_items(self) -> None:
        if not self.allow_sort:
            return
        key = self._sort_key
        self._items.sort(key=key)
        self._refresh_tree()
        self.event_generate("<<ListChanged>>")

    def _sort_key(self, item: Any) -> Any:
        if self.mode == "simple":
            return item
        if self.mode == "kv":
            return (item.get("key"), item.get("value"))
        return (item.get("label"), item.get("value"))

    def dedupe(self) -> None:
        seen: set[Any] = set()
        result: list[Any] = []
        for item in self._items:
            key = self._norm_key(item)
            if key in seen:
                continue
            seen.add(key)
            result.append(item)
        if len(result) != len(self._items):
            self._items = result
            self._refresh_tree()
            self.event_generate("<<ListChanged>>")

    def _norm_key(self, item: Any) -> Any:
        if self.mode == "simple":
            return self._norm(item)
        if self.mode == "kv":
            return self._norm(item.get("key"))
        return self._norm(item.get("value"))

    def _is_duplicate(self, item: Any, items: Sequence[Any] | None = None, idx: int | None = None) -> bool:
        items = items if items is not None else self._items
        key = self._norm_key(item)
        for i, existing in enumerate(items):
            if idx is not None and i == idx:
                continue
            if self._norm_key(existing) == key:
                return True
        return False

    def _on_paste(self) -> None:
        if simpledialog is None:
            return
        text = simpledialog.askstring("Paste", "Enter items, one per line:", parent=self)
        if not text:
            return
        items = [t.strip() for t in text.replace(",", "\n").splitlines()]
        for raw in items:
            if not raw and not self.allow_empty:
                continue
            norm = self._norm(raw)
            ok, _ = self._validate(norm)
            if not ok or (self.unique and self._is_duplicate(norm)):
                continue
            if self.max_items is not None and len(self._items) >= self.max_items:
                break
            self._items.append(norm)
        self._refresh_tree()
        self.event_generate("<<ListChanged>>")

    def _on_import(self) -> None:
        if filedialog is None:
            return
        path = filedialog.askopenfilename(parent=self, filetypes=[("CSV", "*.csv")])
        if not path:
            return

        # ``utf-8-sig`` transparently strips a UTF-8 BOM if present
        with open(path, newline="", encoding="utf-8-sig") as f:

            reader = csv.reader(f)
            rows = list(reader)
        if self.mode != "simple" and rows and len(rows[0]) == len(self._columns):
            rows = rows[1:]
        items: list[Any] = []
        for row in rows:
            if self.mode == "simple":
                items.append(row[0])
            elif self.mode == "kv":
                items.append({"key": row[0], "value": row[1] if len(row) > 1 else ""})
            else:
                items.append({"value": row[0], "label": row[1] if len(row) > 1 else ""})
        self.set_list(items)

    def _on_export(self) -> None:
        if filedialog is None:
            return
        path = filedialog.asksaveasfilename(parent=self, defaultextension=".csv")
        if not path:
            return
        with open(path, "w", newline="", encoding="utf8") as f:
            writer = csv.writer(f)
            if self.mode != "simple":
                writer.writerow([c.heading for c in self._columns])
            for item in self._items:
                writer.writerow(self._item_to_row(item))

    # ------------------------------------------------------------------
    # Prompt helpers
    # ------------------------------------------------------------------
    def _prompt_item(self, initial: Any | None = None) -> Any | None:
        if simpledialog is None:
            return None
        if self.mode == "simple":
            return simpledialog.askstring("Value", "Value:", parent=self, initialvalue=str(initial or ""))
        if self.mode == "kv":
            key = simpledialog.askstring(
                "Key", "Key:", parent=self, initialvalue=str(initial.get("key", "") if isinstance(initial, dict) else "")
            )
            if key is None:
                return None
            val = simpledialog.askstring(
                "Value", "Value:", parent=self, initialvalue=str(initial.get("value", "") if isinstance(initial, dict) else "")
            )
            if val is None:
                return None
            return {"key": key, "value": val}
        value = simpledialog.askstring(
            "Value", "Value:", parent=self, initialvalue=str(initial.get("value", "") if isinstance(initial, dict) else "")
        )
        if value is None:
            return None
        label = simpledialog.askstring(
            "Label", "Label:", parent=self, initialvalue=str(initial.get("label", "") if isinstance(initial, dict) else "")
        )
        if label is None:
            return None
        return {"value": value, "label": label}


class ListEditDialog(tk.Toplevel):  # pragma: no cover - exercised via tk tests
    """Modal dialog embedding :class:`ListEditor`."""

    def __init__(self, parent: tk.Widget, **kwargs: Any) -> None:
        if tk is None:  # pragma: no cover - environment guard
            raise RuntimeError("tkinter is required for ListEditDialog")
        super().__init__(parent)
        use(self)
        palette = get_palette()
        self.title("Edit List")
        self.transient(parent)
        self.grab_set()
        self.resizable(True, True)
        self.configure(bg=palette["bg"])  # type: ignore[call-arg]
        body = ttk.Frame(self, padding=6, style="Card.TFrame")
        body.pack(fill="both", expand=True)
        self.editor = ListEditor(body, **kwargs)
        self.editor.pack(fill="both", expand=True)
        buttons = ttk.Frame(body, style="CardBody.TFrame")
        buttons.pack(fill="x", pady=(6, 0))
        ttk.Button(buttons, text="OK", command=self._on_ok).pack(
            side="right", padx=2
        )
        ttk.Button(buttons, text="Cancel", command=self._on_cancel).pack(
            side="right", padx=2
        )
        self.result: list[Any] | None = None

    def _on_ok(self) -> None:
        # Validate the whole list before accepting
        items = self.editor.get_list()
        for item in items:
            ok, err = self.editor._validate(item)
            if not ok:
                if messagebox is not None:
                    messagebox.showerror("Invalid", err or "invalid value", parent=self)
                self.editor.focus_set()
                self.event_generate("<<ListValidated>>")
                return
        self.result = items
        self.event_generate("<<ListValidated>>")
        self.destroy()

    def _on_cancel(self) -> None:
        self.result = None
        self.destroy()


__all__ = ["ListEditor", "ListEditDialog", "Column"]



from __future__ import annotations

import json
from pathlib import Path
from types import SimpleNamespace

from pysigil import config as cfg
from pysigil.cli import main as cli_main


def _fake_user_dir(tmp_path: Path) -> Path:
    d = tmp_path / "user"
    d.mkdir()
    return d


def _fake_project_root(tmp_path: Path) -> Path:
    root = tmp_path / "proj"
    (root / ".sigil").mkdir(parents=True)
    (root / "pyproject.toml").write_text("[project]\nname='demo'\n")
    return root


def _patch_env(monkeypatch, tmp_path: Path, host: str = "host") -> tuple[Path, Path]:
    user_dir = _fake_user_dir(tmp_path)
    project_root = _fake_project_root(tmp_path)
    monkeypatch.setattr(cfg, "user_config_dir", lambda app: str(user_dir))
    monkeypatch.setattr(cfg, "find_project_root", lambda: project_root)
    monkeypatch.setattr(cfg.socket, "gethostname", lambda: host)
    return user_dir, project_root


def test_precedence(monkeypatch, tmp_path: Path) -> None:
    user_dir, project_root = _patch_env(monkeypatch, tmp_path)
    (user_dir / "pkg").mkdir()
    (user_dir / "pkg" / "settings.ini").write_text("[pkg]\na=1\n")
    (user_dir / "pkg" / "settings-local-host.ini").write_text("[pkg]\na=2\n")
    proj_dir = project_root / ".sigil"
    (proj_dir / "settings.ini").write_text("[pkg]\na=3\n")
    (proj_dir / "settings-local-host.ini").write_text("[pkg]\na=4\n")
    data = cfg.load("pkg")
    assert data == {"a": "4"}

    # switch precedence to user-over-project
    cfg.policy.set_store("user", {("pysigil", "policy"): "user_over_project"})
    try:
        data = cfg.load("pkg")
        assert data == {"a": "2"}
    finally:
        cfg.policy._stores.clear()


def test_invalid_ini_logs_warning(monkeypatch, tmp_path: Path, caplog) -> None:
    user_dir, _ = _patch_env(monkeypatch, tmp_path)
    (user_dir / "pkg").mkdir()
    # write invalid INI content to trigger warning
    (user_dir / "pkg" / "settings.ini").write_text("not an ini")
    with caplog.at_level("WARNING"):
        data = cfg.load("pkg")
    assert data == {}
    assert "Failed to read config" in caplog.text


def _run_cli(args: list[str], capsys) -> SimpleNamespace:
    code = cli_main(args)
    out = capsys.readouterr().out
    return SimpleNamespace(exit_code=code, output=out)


def test_write_policy(monkeypatch, tmp_path: Path, capsys) -> None:
    user_dir, _ = _patch_env(monkeypatch, tmp_path)
    res = _run_cli(["config", "init", "--provider", "user-custom", "--scope", "user"], capsys)
    assert res.exit_code == 0
    assert (user_dir / "user-custom" / "settings-local-host.ini").exists()
    assert not (user_dir / "user-custom" / "settings.ini").exists()
    res = _run_cli(["config", "init", "--provider", "mypkg", "--scope", "user"], capsys)
    assert res.exit_code == 0
    assert (user_dir / "mypkg" / "settings.ini").exists()
    res = _run_cli(["config", "host", "--provider", "mypkg"], capsys)
    assert res.exit_code != 0


def test_host_filtering(monkeypatch, tmp_path: Path) -> None:
    user_dir, project_root = _patch_env(monkeypatch, tmp_path)
    (user_dir / "pkg").mkdir()
    (user_dir / "pkg" / "settings-local-host.ini").write_text("[pkg]\nx=1\n")
    (user_dir / "pkg" / "settings-local-other.ini").write_text("[pkg]\nx=2\n")
    proj_dir = project_root / ".sigil"
    (proj_dir / "settings-local-host.ini").write_text("[pkg]\ny=3\n")
    (proj_dir / "settings-local-other.ini").write_text("[pkg]\ny=4\n")
    data = cfg.load("pkg")
    assert data == {"x": "1", "y": "3"}


def test_gitignore_idempotent(monkeypatch, tmp_path: Path, capsys) -> None:
    _, project_root = _patch_env(monkeypatch, tmp_path)
    _run_cli(["config", "gitignore", "--init", "--auto"], capsys)
    _run_cli(["config", "gitignore", "--init", "--auto"], capsys)
    content = (project_root / ".gitignore").read_text().splitlines()
    assert content == [".sigil/settings-local*"]


def test_cli_roundtrip(monkeypatch, tmp_path: Path, capsys) -> None:
    user_dir, project_root = _patch_env(monkeypatch, tmp_path)
    _run_cli(["config", "init", "--provider", "pkg", "--scope", "user"], capsys)
    _run_cli(["config", "init", "--provider", "pkg", "--scope", "project", "--auto"], capsys)
    (user_dir / "pkg" / "settings.ini").write_text("[pkg]\nkey=user\n")
    proj_dir = project_root / ".sigil"
    (proj_dir / "settings.ini").write_text("[pkg]\nkey=project\n")
    res = _run_cli(["config", "show", "--provider", "pkg", "--as", "json", "--auto"], capsys)
    assert res.exit_code == 0
    assert json.loads(res.output) == {"key": "project"}



from __future__ import annotations

import configparser
import re
from dataclasses import dataclass
from pathlib import Path

import tomlkit

from .paths import user_config_dir

try:  # pragma: no cover - fallback when setuptools is missing
    from setuptools import PackageFinder  # type: ignore
except Exception:  # pragma: no cover - defensive
    PackageFinder = None  # type: ignore[assignment]

# ---------------------------------------------------------------------------
# Dev links registry
# ---------------------------------------------------------------------------


class DevLinkError(Exception):
    """Raised when development link operations fail."""


@dataclass
class DevLink:
    provider_id: str
    defaults_path: Path
    link_path: Path


_PEP503_RE = re.compile(r"[-_.]+")
_VALID_RE = re.compile(r"^[a-z0-9-]+$")


def normalize_provider_id(name: str) -> str:
    """PEP 503 normalize ``name``.

    ``name`` is normalised to lowercase with hyphens and must not contain path
    separators or ``..`` segments.
    """

    stripped = name.strip().lower()
    norm = _PEP503_RE.sub("-", stripped)
    if (
        "/" in stripped
        or "\\" in stripped
        or ".." in stripped
        or not _VALID_RE.fullmatch(norm)
    ):
        raise ValueError(f"invalid provider id: {name!r}")
    return norm


def _dev_dir() -> Path:
    return Path(user_config_dir("sigil")) / "dev"


def link(provider: str, defaults_path: Path, *, validate: bool = True) -> DevLink:
    pid = normalize_provider_id(provider)
    defaults_path = Path(defaults_path).expanduser().resolve()
    if not defaults_path.is_file():
        raise DevLinkError(f"Defaults file not found: {defaults_path}")
    if defaults_path.name != "settings.ini" or defaults_path.parent.name != ".sigil":
        raise DevLinkError("Defaults file must be inside a '.sigil' directory")
    if validate:
        validate_defaults_file(defaults_path, pid)
    dir_ = _dev_dir()
    dir_.mkdir(parents=True, exist_ok=True)
    link_path = dir_ / f"{pid}.ini"
    cfg = configparser.ConfigParser()
    cfg["link"] = {"defaults": str(defaults_path)}
    with link_path.open("w") as f:
        cfg.write(f)
    return DevLink(pid, defaults_path, link_path)


def unlink(provider: str) -> bool:
    pid = normalize_provider_id(provider)
    link_path = _dev_dir() / f"{pid}.ini"
    if link_path.exists():
        link_path.unlink()
        return True
    return False


def get(provider: str) -> DevLink | None:
    pid = normalize_provider_id(provider)
    link_path = _dev_dir() / f"{pid}.ini"
    if not link_path.is_file():
        return None
    cfg = configparser.ConfigParser()
    try:
        cfg.read(link_path)
    except Exception:
        return None
    defaults = cfg.get("link", "defaults", fallback=None)
    if not defaults:
        return None
    return DevLink(pid, Path(defaults), link_path)


def list_links(must_exist_on_disk: bool = False) -> dict[str, Path]:
    dir_ = _dev_dir()
    if not dir_.is_dir():
        return {}
    result: dict[str, Path] = {}
    for file in dir_.glob("*.ini"):
        dl = get(file.stem)
        if dl is None:
            continue
        if must_exist_on_disk and not dl.defaults_path.exists():
            continue
        result[dl.provider_id] = dl.defaults_path
    return result


# ---------------------------------------------------------------------------
# Defaults validation
# ---------------------------------------------------------------------------


class DefaultsValidationError(Exception):
    """Raised when a defaults file is invalid."""


_KEY_RE = re.compile(r"^[a-z0-9][a-z0-9_]*(\.[a-z0-9_]+)*$")


def validate_defaults_file(path: Path, provider_id: str) -> None:
    """Validate ``path`` for ``provider_id``.

    The file must contain a ``[<provider_id>]`` section and all keys in that
    section must be dotted names using lowercase letters, digits, and optional
    underscores. Keys must start with a letter or digit.
    """

    path = Path(path).expanduser().resolve()
    if not path.is_file():
        raise DefaultsValidationError("defaults file not found")

    parser = configparser.ConfigParser()
    try:
        parser.read(path)
    except Exception as exc:  # pragma: no cover - defensive
        raise DefaultsValidationError(str(exc)) from exc

    section = provider_id
    if section not in parser:
        raise DefaultsValidationError("missing provider section")
    for key in parser[section]:
        if not _KEY_RE.match(key):
            raise DefaultsValidationError(f"invalid key: {key}")


# ---------------------------------------------------------------------------
# pyproject.toml helpers
# ---------------------------------------------------------------------------


def import_package_from(ini_path: Path) -> str:
    """Infer the import package name from ``ini_path``."""

    ini_path = Path(ini_path).resolve()
    if PackageFinder is not None:
        for ancestor in ini_path.parents:
            for base in [ancestor / "src", ancestor]:
                if not base.is_dir():
                    continue
                for pkg in PackageFinder.find(where=[str(base)]):
                    pkg_dir = base / Path(pkg.replace(".", "/"))
                    if pkg_dir / ".sigil" / "settings.ini" == ini_path:
                        return pkg
    return ini_path.parent.parent.name


def ensure_sigil_package_data(project_root: Path, package: str) -> None:
    """Ensure ``pyproject.toml`` includes ``.sigil`` package data for *package*.

    If ``pyproject.toml`` is missing or can't be parsed, this is a no-op.
    The function sets ``[tool.setuptools.package-data][<package>]`` to
    ``[".sigil/*"]``. Other packages' configuration is preserved.
    """

    ppt = project_root / "pyproject.toml"
    if not ppt.is_file():
        print("Error locating pyproject.toml")
        return

    try:  # pragma: no cover - defensive
        doc = tomlkit.parse(ppt.read_text(encoding="utf-8"))
    except Exception:  # pragma: no cover - defensive
        print("Error reading pyproject.toml")
        return

    tool = doc.setdefault("tool", tomlkit.table())
    st = tool.setdefault("setuptools", tomlkit.table())
    pd = st.setdefault("package-data", tomlkit.table())

    arr = pd.get(package)
    desired = [".sigil/*"]
    current = [str(x) for x in arr] if arr is not None else None
    if current != desired:
        new_arr = tomlkit.array().multiline(True)
        for item in desired:
            new_arr.append(item)
        pd[package] = new_arr
        ppt.write_text(tomlkit.dumps(doc), encoding="utf-8")



"""Framework agnostic GUI core.

The :mod:`pysigil.ui.core` module exposes the minimal pieces required to
build a graphical user interface on top of :mod:`pysigil.api`.  The code
here purposefully knows nothing about any widget toolkit; instead it is
responsible for keeping track of application state, delegating all heavy
lifting to :mod:`pysigil.api` and notifying interested parties about
state changes via a tiny callback based event system.

The intention is that concrete front-ends – for instance a tkinter based
one – bind to the callbacks exposed here.  Replacing the front-end should
only require reimplementing the view layer while this module can be
reused unchanged.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Literal, Protocol
from concurrent.futures import ThreadPoolExecutor, Future
import threading
from pathlib import Path

from .. import api, config, paths

# ---------------------------------------------------------------------------
# Application state
# ---------------------------------------------------------------------------


@dataclass(slots=True)
class AppState:
    """In-memory representation of the UI state."""

    provider_id: str | None = None
    provider_info: api.ProviderInfo | None = None
    fields: list[api.FieldInfo] = field(default_factory=list)
    values: Dict[str, api.ValueInfo] = field(default_factory=dict)
    active_tab: Literal["overview", "user", "project"] = "overview"
    active_scope: Literal["user", "project"] = "user"
    host_id: str = config.host_id()
    project_root: Path | None = None
    is_dirty_spec: bool = False
    is_dirty_values: bool = False
    project_detected: bool = True
    author_mode: bool = False


# ---------------------------------------------------------------------------
# Event handling
# ---------------------------------------------------------------------------


class EventBus:
    """Simple callback based pub/sub system."""

    def __init__(self) -> None:
        self.on_state_changed: list[Callable[[AppState], None]] = []
        self.on_error: list[Callable[[str], None]] = []
        self.on_toast: list[Callable[[str, str], None]] = []
        self.on_confirm: list[Callable[[str, str], bool]] = []
        self.on_progress: list[Callable[[bool], None]] = []

    # Emit helpers -----------------------------------------------------
    def emit_state(self, state: AppState) -> None:
        for cb in list(self.on_state_changed):
            cb(state)

    def emit_error(self, msg: str) -> None:
        for cb in list(self.on_error):
            cb(msg)

    def emit_toast(self, msg: str, level: str = "info") -> None:
        for cb in list(self.on_toast):
            cb(msg, level)

    def emit_progress(self, started: bool) -> None:
        for cb in list(self.on_progress):
            cb(started)


# ---------------------------------------------------------------------------
# Service façade (thin wrapper over :mod:`pysigil.api`)
# ---------------------------------------------------------------------------


class ProvidersService:
    """Lightweight façade over :mod:`pysigil.api`.

    The service performs no caching; callers are expected to cache data in
    :class:`AppState` if necessary.
    """

    def __init__(self, *, author_mode: bool = False) -> None:
        self.author_mode = author_mode

    def list_providers(self) -> list[str]:
        return api.providers()

    def select_provider(self, pid: str) -> api.ProviderInfo:
        return api.get_provider(pid)

    def get_fields(self, pid: str) -> list[api.FieldInfo]:
        return api.handle(pid).fields()

    def get_effective(self, pid: str) -> Dict[str, api.ValueInfo]:
        return api.handle(pid).effective()

    def get_layers(self, pid: str) -> Dict[str, Dict[str, api.ValueInfo | None]]:
        return api.handle(pid).layers()

    def add_field(
        self,
        pid: str,
        key: str,
        type: str,
        *,
        label: str | None = None,
        description_short: str | None = None,
        description: str | None = None,
        options: Dict[str, Any] | None = None,
        init_scope: Literal["user", "project"] | None = "user",
    ) -> api.FieldInfo:
        return api.handle(pid).add_field(
            key,
            type,
            label=label,
            description_short=description_short,
            description=description,
            options=options,
            init_scope=init_scope,
        )

    def set_value(
        self,
        pid: str,
        key: str,
        value: Any,
        *,
        scope: Literal["user", "project", "default"] = "user",
    ) -> None:
        if scope == "default":
            if not self.author_mode:
                raise PermissionError("default scope is read-only")
            api.handle(pid)._manager().set(  # type: ignore[attr-defined]
                key, value, scope="default"
            )
            return
        api.handle(pid).set(key, value, scope=scope)

    def clear_value(
        self,
        pid: str,
        key: str,
        *,
        scope: Literal["user", "project", "default"] = "user",
    ) -> None:
        if scope == "default":
            if not self.author_mode:
                raise PermissionError("default scope is read-only")
            api.handle(pid)._manager().clear(  # type: ignore[attr-defined]
                key, scope="default"
            )
            return
        api.handle(pid).clear(key, scope=scope)

    def init(self, pid: str, scope: Literal["user", "project"]) -> None:
        api.handle(pid).init(scope)

    def edit_field(
        self,
        pid: str,
        key: str,
        *,
        new_key: str | None = None,
        new_type: str | None = None,
        label: str | None = None,
        description_short: str | None = None,
        description: str | None = None,
        options: Dict[str, Any] | None = None,
        on_type_change: Literal["convert", "clear"] = "convert",
        migrate_scopes: tuple[Literal["user", "project"], ...] = ("user",),
    ) -> api.FieldInfo:
        return api.handle(pid).edit_field(
            key,
            new_key=new_key,
            new_type=new_type,
            label=label,
            description_short=description_short,
            description=description,
            options=options,
            on_type_change=on_type_change,
            migrate_scopes=migrate_scopes,
        )

    def delete_field(
        self,
        pid: str,
        key: str,
        *,
        remove_values: bool = False,
        scopes: tuple[Literal["user", "project"], ...] = ("user",),
    ) -> None:
        api.handle(pid).delete_field(key, remove_values=remove_values, scopes=scopes)

    def adopt_untracked(self, pid: str, mapping: Dict[str, str]) -> list[api.FieldInfo]:
        return api.handle(pid).adopt_untracked(mapping)

    def open_folder(self, pid: str, scope: Literal["user", "project"]) -> Path:
        path = config.open_scope(pid, scope)
        self._open_path(path)
        return path

    def open_file(self, pid: str, scope: Literal["user", "project"]) -> Path:
        path = config.init_config(pid, scope)
        self._open_path(path)
        return path

    def ensure_gitignore(self) -> Path:
        return config.ensure_gitignore()

    # -- helpers -----------------------------------------------------
    @staticmethod
    def _open_path(path: Path) -> None:
        try:  # pragma: no cover - platform specific
            import os, subprocess, sys

            if sys.platform.startswith("darwin"):
                subprocess.run(["open", str(path)], check=False)
            elif os.name == "nt":
                os.startfile(str(path))  # type: ignore[attr-defined]
            else:
                subprocess.run(["xdg-open", str(path)], check=False)
        except Exception:
            pass

    def export_spec(self, pid: str) -> None:  # pragma: no cover - thin pass through
        api.handle(pid).export_spec()

    def reload_spec(self, pid: str) -> api.ProviderInfo:
        return api.handle(pid).info()


# ---------------------------------------------------------------------------
# App core / command helpers
# ---------------------------------------------------------------------------


class AppCore:
    """Mediator between the view and :class:`ProvidersService`.

    It keeps the current :class:`AppState` and exposes a small selection of
    command methods which operate on the service.  All operations involving
    the service are executed in a thread pool in order to avoid blocking the
    UI thread in front-ends such as tkinter.
    """

    def __init__(
        self,
        service: ProvidersService | None = None,
        *,
        executor: ThreadPoolExecutor | None = None,
        author_mode: bool = False,
    ) -> None:
        self.state = AppState(author_mode=author_mode)
        self.service = service or ProvidersService(author_mode=author_mode)
        self.events = EventBus()
        self.orchestrator = api._ORCH
        self._executor = executor or ThreadPoolExecutor(max_workers=4)
        self._lock = threading.Lock()

    # --- concurrency -------------------------------------------------
    def run_async(self, fn: Callable[[], Any]) -> Future[Any]:
        """Run ``fn`` in the thread pool and return the future."""

        def runner() -> Any:
            try:
                self.events.emit_progress(True)
                return fn()
            finally:
                self.events.emit_progress(False)

        return self._executor.submit(runner)

    # --- commands ----------------------------------------------------
    def select_provider(self, pid: str) -> Future[None]:
        """Load provider metadata and effective values."""

        def _task() -> None:
            info = self.service.select_provider(pid)
            fields = self.service.get_fields(pid)
            values = self.service.get_effective(pid)
            try:
                root = paths.project_root()
                detected = True
            except Exception:
                root = None
                detected = False
            with self._lock:
                self.state.provider_id = pid
                self.state.provider_info = info
                self.state.fields = fields
                self.state.values = values
                self.state.project_root = root
                self.state.project_detected = detected
            self.events.emit_state(self.state)

        return self.run_async(_task)

    def refresh(self) -> Future[None]:
        """Refresh fields and values for the current provider."""

        pid = self.state.provider_id
        if pid is None:  # nothing to do
            future: Future[None] = Future()
            future.set_result(None)
            return future

        def _task() -> None:
            info = self.service.reload_spec(pid)
            fields = self.service.get_fields(pid)
            values = self.service.get_effective(pid)
            try:
                root = paths.project_root()
                detected = True
            except Exception:
                root = None
                detected = False
            with self._lock:
                self.state.provider_info = info
                self.state.fields = fields
                self.state.values = values
                self.state.project_root = root
                self.state.project_detected = detected
            self.events.emit_state(self.state)

        return self.run_async(_task)

    def save_value(
        self, key: str, value: Any, *, scope: Literal["user", "project"] = "user"
    ) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                self.service.set_value(pid, key, value, scope=scope)
                val = self.service.get_effective(pid).get(key)
                if val is not None:
                    with self._lock:
                        self.state.values[key] = val
                self.events.emit_toast("Saved", "info")
                self.events.emit_state(self.state)
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def clear_value(
        self, key: str, *, scope: Literal["user", "project"] = "user"
    ) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                self.service.clear_value(pid, key, scope=scope)
                with self._lock:
                    self.state.values.pop(key, None)
                self.events.emit_toast("Cleared", "info")
                self.events.emit_state(self.state)
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def init_scope(self, scope: Literal["user", "project"]) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                self.service.init(pid, scope)
                self.events.emit_toast(f"Initialized {scope} scope", "info")
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def open_folder(self, scope: Literal["user", "project"]) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                path = self.service.open_folder(pid, scope)
                self.events.emit_toast(str(path), "info")
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def open_file(self, scope: Literal["user", "project"]) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                path = self.service.open_file(pid, scope)
                self.events.emit_toast(str(path), "info")
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def add_gitignore(self) -> Future[None]:
        def _task() -> None:
            try:
                path = self.service.ensure_gitignore()
                self.events.emit_toast(str(path), "info")
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def set_active_scope(self, scope: Literal["user", "project"]) -> None:
        with self._lock:
            self.state.active_scope = scope
        self.events.emit_state(self.state)

    def add_field(
        self,
        key: str,
        type: str,
        *,
        label: str | None = None,
        description: str | None = None,
        init_scope: Literal["user", "project"] | None = "user",
    ) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                self.service.add_field(
                    pid,
                    key,
                    type,
                    label=label,
                    description=description,
                    init_scope=init_scope,
                )
                self.refresh()
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def edit_field(
        self,
        key: str,
        *,
        new_key: str | None = None,
        new_type: str | None = None,
        label: str | None = None,
        description: str | None = None,
        on_type_change: Literal["convert", "clear"] = "convert",
        migrate_scopes: tuple[Literal["user", "project"], ...] = ("user",),
    ) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                self.service.edit_field(
                    pid,
                    key,
                    new_key=new_key,
                    new_type=new_type,
                    label=label,
                    description=description,
                    on_type_change=on_type_change,
                    migrate_scopes=migrate_scopes,
                )
                self.refresh()
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def delete_field(
        self,
        key: str,
        *,
        remove_values: bool = False,
        scopes: tuple[Literal["user", "project"], ...] = ("user",),
    ) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                self.service.delete_field(
                    pid, key, remove_values=remove_values, scopes=scopes
                )
                self.refresh()
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)

    def adopt_untracked(self, mapping: Dict[str, str]) -> Future[None]:
        pid = self.state.provider_id
        if pid is None:
            raise RuntimeError("no provider selected")

        def _task() -> None:
            try:
                self.service.adopt_untracked(pid, mapping)
                self.refresh()
            except Exception as exc:
                self.events.emit_error(str(exc))
                self.events.emit_toast(str(exc), "error")

        return self.run_async(_task)


__all__ = [
    "AppCore",
    "AppState",
    "EventBus",
    "ProvidersService",
]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Iterator, List, MutableMapping, Tuple

from .errors import SigilWriteError, UnknownScopeError
from .root import ProjectRootNotFoundError

# Predefined defaults for the core scope
CORE_DEFAULTS = {"pysigil": {"policy": "project_over_user"}}

# Precedence orders from highest to lowest scope
PRECEDENCE_USER_WINS: Tuple[str, ...] = (
    "env",
    "user-local",
    "user",
    "project-local",
    "project",
    "default",
    "core",
)
PRECEDENCE_PROJECT_WINS: Tuple[str, ...] = (
    "env",
    "project-local",
    "project",
    "user-local",
    "user",
    "default",
    "core",
)


@dataclass(frozen=True)
class Scope:
    """Representation of a configuration scope."""

    name: str
    writable: bool = False
    machine: bool = False


class ScopePolicy:
    """Manager providing policy information for configuration scopes."""

    def __init__(self, scopes: Iterable[Scope]):
        self._scopes: List[Scope] = list(scopes)
        self._by_name = {s.name: s for s in self._scopes}
        self._stores: dict[str, MutableMapping] = {}
        self._machine = {s.name for s in self._scopes if s.machine}

    @property
    def scopes(self) -> List[str]:
        """Ordered list of known scope names."""
        return [s.name for s in self._scopes]

    def precedence(self, *, read: bool = False) -> Tuple[str, ...]:
        """Return the precedence order.

        When ``read`` is ``True`` the current policy is looked up from the
        stored configuration.  If no policy is configured the project-over-user
        order is assumed.
        """

        if not read:
            return PRECEDENCE_PROJECT_WINS

        key = ("pysigil", "policy")
        for scope in PRECEDENCE_PROJECT_WINS:
            store = self._stores.get(scope)
            if store is None:
                continue
            mode = store.get(key)
            if mode == "user_over_project":
                return PRECEDENCE_USER_WINS
            if mode == "project_over_user":
                return PRECEDENCE_PROJECT_WINS
        return PRECEDENCE_PROJECT_WINS

    def allows(self, scope: str) -> bool:
        """Return ``True`` if *scope* is writable."""
        info = self._by_name.get(scope)
        if info is None:
            raise UnknownScopeError(scope)
        return info.writable

    # ----- helpers for Sigil integration -----

    def clone(self, *, default_writable: bool | None = None) -> "ScopePolicy":
        """Return a shallow copy optionally adjusting default scope writability."""

        scopes = [Scope(s.name, s.writable) for s in self._scopes]
        if default_writable is not None:
            scopes = [
                Scope(s.name, default_writable if s.name == "default" else s.writable)
                for s in scopes
            ]
        return ScopePolicy(scopes)

    def set_store(self, scope: str, store: MutableMapping) -> None:
        """Associate *store* with *scope* for later retrieval."""

        if scope not in self._by_name:
            raise UnknownScopeError(scope)
        self._stores[scope] = store

    def get_store(self, scope: str) -> MutableMapping:
        """Return the mapping associated with *scope*."""

        try:
            return self._stores[scope]
        except KeyError as exc:
            raise UnknownScopeError(scope) from exc

    def iter_scopes(self, *, read: bool = False) -> Iterator[str]:
        """Yield scope names in precedence order."""

        # For now reading order is identical to definition order
        for s in self._scopes:
            yield s.name

    def machine_scopes(self) -> List[str]:
        """Return scopes that are machine-specific."""

        return list(self._machine)

    def path(self, scope: str, provider_id: str, *, auto: bool = False) -> Path:
        """Return configuration path for *scope* and *provider_id*.

        Directories are created as needed and writes are only permitted for
        scopes marked writable in the policy.
        """

        if not self.allows(scope):
            raise SigilWriteError(f"Scope '{scope}' is read-only")

        from . import config
        from .authoring import normalize_provider_id

        pid = normalize_provider_id(provider_id)
        if scope in {"user", "user-local"}:
            base = Path(config.user_config_dir("sigil")) / pid
        else:
            root = config._project_dir(auto)
            if root is None:
                raise ProjectRootNotFoundError("No project root found")
            base = root / ".sigil"
        base.mkdir(parents=True, exist_ok=True)
        if scope in self._machine or pid == "user-custom":
            return base / f"settings-local-{config.host_id()}.ini"
        return base / "settings.ini"


# Known scopes ordered from lowest to highest precedence
_DEFAULT_SCOPES = [
    Scope("core", writable=False),
    Scope("default", writable=False),
    Scope("project", writable=True),
    Scope("project-local", writable=True, machine=True),
    Scope("user", writable=True),
    Scope("user-local", writable=True, machine=True),
    Scope("env", writable=False),
]

policy = ScopePolicy(_DEFAULT_SCOPES)

__all__ = [
    "Scope",
    "ScopePolicy",
    "policy",
    "CORE_DEFAULTS",
    "PRECEDENCE_USER_WINS",
    "PRECEDENCE_PROJECT_WINS",
]



from pysigil import Sigil
from pysigil.policy import policy


def test_defaults_section_normalized(tmp_path):
    defaults_dir = tmp_path / ".sigil"
    defaults_dir.mkdir()
    defaults_file = defaults_dir / "settings.ini"
    defaults_file.write_text("[sigil_dummy]\nfoo=bar\n", encoding="utf-8")

    sig = Sigil("sigil-dummy", default_path=defaults_file, user_scope=tmp_path / "user.ini", policy=policy)
    assert sig.get_pref("foo") == "bar"
    assert ("foo",) in sig.list_keys("default")


def test_env_prefix_normalized(tmp_path, monkeypatch):
    monkeypatch.setenv("SIGIL_SIGIL_DUMMY_FOO", "42")
    sig = Sigil("sigil-dummy", user_scope=tmp_path / "user.ini", policy=policy)
    assert sig.get_pref("foo") == 42



from pathlib import Path

import pytest

from pysigil.root import (
    ProjectRootNotFoundWithSuggestionsError,
    find_project_root,
)


def mkfile(base: Path, rel: str, content: str = "x") -> Path:
    p = base / rel
    p.parent.mkdir(parents=True, exist_ok=True)
    p.write_text(content, encoding="utf-8")
    return p

def mkdir(base: Path, rel: str) -> Path:
    p = base / rel
    p.mkdir(parents=True, exist_ok=True)
    return p

def test_vscode_core_files_any_one_is_enough(tmp_path: Path):
    proj = tmp_path / "vs"
    mkfile(proj, ".vscode/launch.json", "{}")
    assert find_project_root(start=proj / "a") == proj

def test_idea_core_files_any_one_is_enough(tmp_path: Path):
    proj = tmp_path / "ij"
    mkfile(proj, ".idea/modules.xml", "<modules/>")
    assert find_project_root(start=proj / "a") == proj

def test_ide_guard_prefers_python_if_present(tmp_path: Path):
    proj = tmp_path / "both"
    mkfile(proj, ".vscode/settings.json", "{}")
    mkfile(proj, "pyproject.toml", "")
    # Should return immediately at proj due to Tier1; test ensures no guard interference
    assert find_project_root(start=proj / "deep") == proj

def test_empty_vscode_and_no_co_signals_gives_suggestions(tmp_path: Path):
    proj = tmp_path / "empty"
    mkdir(proj, ".vscode")
    with pytest.raises(ProjectRootNotFoundWithSuggestionsError):
        find_project_root(start=proj / "child")



from __future__ import annotations

from ..settings_metadata import TYPE_REGISTRY


def parse_field_value(type_name: str, raw: object) -> object:
    """Parse *raw* text into a Python value for ``type_name``.

    ``raw`` is typically the content of a text entry widget.  The function
    accepts ``1``/``0`` for booleans in addition to ``true``/``false`` to mirror
    the behaviour of the Tk dialogs.
    """
    if raw is None:
        return None
    if isinstance(raw, str):
        text = raw.strip()
        if type_name == "boolean":
            lower = text.lower()
            if lower == "1":
                text = "true"
            elif lower == "0":
                text = "false"
        adapter = TYPE_REGISTRY[type_name].adapter
        return adapter.parse(text)
    return raw



from __future__ import annotations

import os
from pathlib import Path

from pysigil import Sigil
from pysigil.policy import policy


def test_project_context_updates_local_path(tmp_path: Path) -> None:
    root_a = tmp_path / "a"
    root_b = tmp_path / "b"
    for root in (root_a, root_b):
        (root / ".sigil").mkdir(parents=True)
        (root / ".sigil-root").touch()

    cwd = os.getcwd()
    os.chdir(root_a)
    try:
        s = Sigil("demo", user_scope=tmp_path / "user", policy=policy)
        host = s._host
        old_local = root_a / ".sigil" / f"settings-local-{host}.ini"

        with s.project(root_b / ".sigil" / "settings.ini"):
            s.set_pref("foo", "bar", scope="project-local")
            expected = root_b / ".sigil" / f"settings-local-{host}.ini"
            assert s.project_local_path == expected
            assert s.path_for_scope("project-local") == expected
            assert expected.exists()
            assert not old_local.exists()
    finally:
        os.chdir(cwd)
    # path restored
    assert s.project_local_path == old_local



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, field as dataclass_field
from pathlib import Path
from typing import Any, Literal
import os

from .errors import (
    ConflictError,
    DuplicateFieldError,
    DuplicateProviderError,
    IOFailureError,
    PolicyError,
    SigilError,
    UnknownFieldError,
    UnknownProviderError,
    ValidationError,
)
from .orchestrator import Orchestrator
from .policy import policy
from .settings_metadata import FieldSpec, FieldValue, ProviderSpec

__all__ = [
    "SigilError",
    "UnknownProviderError",
    "UnknownFieldError",
    "DuplicateFieldError",
    "ValidationError",
    "PolicyError",
    "ConflictError",
    "IOFailureError",
    "FieldInfo",
    "ValueInfo",
    "ProviderInfo",
    "providers",
    "get_provider",
    "handle",
    "register_provider",
    "ProviderHandle",
]


# ---------------------------------------------------------------------------
# dataclasses returned to callers
# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class FieldInfo:
    key: str
    type: str
    label: str | None
    description_short: str | None
    description: str | None
    options: dict[str, Any] = dataclass_field(default_factory=dict)
    section: str | None = None
    order: int | None = None


@dataclass(frozen=True)
class ValueInfo:
    value: Any | None
    source: Literal["user", "user-local", "project", "project-local", "env"] | None
    raw: str | None


@dataclass(frozen=True)
class ProviderInfo:
    provider_id: str
    title: str | None
    description: str | None
    fields: list[FieldInfo]
    sections_order: list[str] | None = None
    sections_collapsed: list[str] | None = None


def _field_info(spec: FieldSpec) -> FieldInfo:
    return FieldInfo(
        key=spec.key,
        type=spec.type,
        label=spec.label,
        description_short=spec.description_short,
        description=spec.description,
        options=spec.options,
        section=spec.section,
        order=spec.order,
    )


def _provider_info(spec: ProviderSpec) -> ProviderInfo:
    return ProviderInfo(
        provider_id=spec.provider_id,
        title=spec.title,
        description=spec.description,
        fields=[_field_info(f) for f in spec.fields],
        sections_order=list(spec.sections_order)
        if spec.sections_order is not None
        else None,
        sections_collapsed=list(spec.sections_collapsed)
        if spec.sections_collapsed is not None
        else None,
    )


def _value_info(val: FieldValue) -> ValueInfo:
    return ValueInfo(value=val.value, source=val.source, raw=val.raw)


# ---------------------------------------------------------------------------
# orchestrator wiring
# ---------------------------------------------------------------------------

_ORCH = Orchestrator()


# ---------------------------------------------------------------------------
# top level API
# ---------------------------------------------------------------------------

def providers() -> list[str]:
    """List all registered provider ids."""
    return _ORCH.list_providers()


def get_provider(provider_id: str) -> ProviderInfo:
    spec = _ORCH.reload_spec(provider_id)
    return _provider_info(spec)


def handle(provider_id: str) -> ProviderHandle:
    get_provider(provider_id)  # ensure existence
    return ProviderHandle(provider_id)


def register_provider(
    provider_id: str,
    *,
    title: str | None = None,
    description: str | None = None,
) -> ProviderInfo:
    try:
        spec = _ORCH.register_provider(
            provider_id, title=title, description=description
        )
    except DuplicateProviderError:
        spec = _ORCH.reload_spec(provider_id)
    return _provider_info(spec)


# ---------------------------------------------------------------------------
# provider bound façade
# ---------------------------------------------------------------------------


class ProviderHandle:
    provider_id: str

    def __init__(self, provider_id: str) -> None:
        self.provider_id = provider_id

    def _manager(self):  # pragma: no cover - simple wrapper
        return _ORCH._manager(self.provider_id)  # type: ignore[attr-defined]

    def _env_key(self, key: str) -> str:
        base = self.provider_id.upper().replace("-", "_")
        return f"SIGIL_{base}_{key.replace('.', '_').replace('-', '_').upper()}"

    @staticmethod
    def _normalize_scope(scope: str) -> str:
        return "env" if scope == "environment" else scope

    # -------- Metadata (field catalog) --------
    def info(self) -> ProviderInfo:
        return get_provider(self.provider_id)

    def fields(self) -> list[FieldInfo]:
        return self.info().fields

    def add_field(
        self,
        key: str,
        type: str,
        *,
        label: str | None = None,
        description_short: str | None = None,
        description: str | None = None,
        options: Mapping[str, Any] | None = None,
        section: str | None = None,
        order: int | None = None,
        init_scope: Literal[
            "user", "user-local", "project", "project-local", "environment"
        ] | None = "user",
    ) -> FieldInfo:
        try:
            field = _ORCH.add_field(
                self.provider_id,
                key=key,
                type=type,
                label=label,
                description_short=description_short,
                description=description,
                options=dict(options) if options is not None else None,
                section=section,
                order=order,
            )
        except DuplicateFieldError as exc:
            raise DuplicateFieldError(key) from exc
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc
        if init_scope is not None and init_scope != "environment":
            try:
                self._manager().init(init_scope)
            except PolicyError as exc:  # pragma: no cover - defensive
                raise PolicyError(str(exc)) from exc
        return _field_info(field)

    def edit_field(
        self,
        key: str,
        *,
        new_key: str | None = None,
        new_type: str | None = None,
        label: str | None = None,
        description_short: str | None = None,
        description: str | None = None,
        options: Mapping[str, Any] | None = None,
        section: str | None = None,
        order: int | None = None,
        on_type_change: Literal["convert", "clear"] = "convert",
        migrate_scopes: tuple[
            Literal["user", "user-local", "project", "project-local"], ...
        ] = ("user",),
    ) -> FieldInfo:
        try:
            field = _ORCH.edit_field(
                self.provider_id,
                key,
                new_key=new_key,
                new_type=new_type,
                label=label,
                description_short=description_short,
                description=description,
                options=dict(options) if options is not None else None,
                section=section,
                order=order,
                on_type_change=on_type_change,
            )
        except UnknownFieldError as exc:
            raise UnknownFieldError(key) from exc
        except DuplicateFieldError as exc:
            raise DuplicateFieldError(str(exc)) from exc
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc
        return _field_info(field)

    def delete_field(
        self,
        key: str,
        *,
        remove_values: bool = False,
        scopes: tuple[
            Literal["user", "user-local", "project", "project-local"], ...
        ] = ("user", "project"),
    ) -> None:
        try:
            _ORCH.delete_field(
                self.provider_id,
                key,
                remove_values=remove_values,
                scopes=scopes,
            )
        except UnknownFieldError as exc:
            raise UnknownFieldError(key) from exc
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc

    def get_sections_order(self) -> list[str] | None:
        return _ORCH.get_sections_order(self.provider_id)

    def set_sections_order(self, seq: list[str]) -> None:
        _ORCH.set_sections_order(self.provider_id, list(seq))

    def get_sections_collapsed(self) -> list[str] | None:
        return _ORCH.get_sections_collapsed(self.provider_id)

    def set_sections_collapsed(self, seq: list[str]) -> None:
        _ORCH.set_sections_collapsed(self.provider_id, list(seq))

    def patch_fields(self, updates: list[dict[str, object]]) -> None:
        _ORCH.patch_fields(self.provider_id, updates)

    # Untracked (manual INI edits)
    def untracked_keys(self) -> list[str]:
        return _ORCH.find_untracked_keys(self.provider_id)

    def adopt_untracked(self, mapping: Mapping[str, str]) -> list[FieldInfo]:
        specs = _ORCH.adopt_untracked(self.provider_id, dict(mapping))
        return [_field_info(s) for s in specs]

    # -------- Values (typed + scope aware) --------
    def effective(self) -> dict[str, ValueInfo]:
        values = _ORCH.get_effective(self.provider_id)
        return {k: _value_info(v) for k, v in values.items()}

    def layers(self) -> dict[str, dict[str, ValueInfo | None]]:
        """Return values for all scopes for this provider."""
        layers = _ORCH.get_layers(self.provider_id)
        result: dict[str, dict[str, ValueInfo | None]] = {}
        for key, per_scope in layers.items():
            result[key] = {
                scope: (_value_info(v) if v is not None else None)
                for scope, v in per_scope.items()
            }
        return result

    def get(self, key: str) -> ValueInfo:
        eff = self.effective()
        if key not in eff:
            raise UnknownFieldError(key)
        return eff[key]

    def set(
        self,
        key: str,
        value: object,
        *,
        scope: Literal[
            "user", "user-local", "project", "project-local", "environment"
        ] = "user",
    ) -> None:
        scope = self._normalize_scope(scope)
        if scope == "env":
            os.environ[self._env_key(key)] = str(value)
            return
        try:
            _ORCH.set_value(self.provider_id, key, value, scope=scope)
        except ValidationError as exc:
            raise ValidationError(str(exc)) from exc
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc
        except KeyError as exc:  # propagated by provider manager
            raise UnknownFieldError(key) from exc

    def clear(
        self,
        key: str,
        *,
        scope: Literal[
            "user", "user-local", "project", "project-local", "environment"
        ] = "user",
    ) -> None:
        scope = self._normalize_scope(scope)
        if scope == "env":
            os.environ.pop(self._env_key(key), None)
            return
        try:
            _ORCH.clear_value(self.provider_id, key, scope=scope)
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc
        except KeyError as exc:
            raise UnknownFieldError(key) from exc

    def set_many(
        self,
        updates: Mapping[str, object],
        *,
        scope: Literal[
            "user", "user-local", "project", "project-local", "environment"
        ] = "user",
        atomic: bool = True,
    ) -> None:
        """Set multiple configuration values.

        With ``atomic=True`` (the default) updates are staged and committed
        only if all validations and writes succeed.
        """
        scope = self._normalize_scope(scope)
        if scope == "env":
            for k, v in updates.items():
                os.environ[self._env_key(k)] = str(v)
            return
        try:
            _ORCH.set_many(
                self.provider_id, dict(updates), scope=scope, atomic=atomic
            )
        except ValidationError as exc:
            raise ValidationError(str(exc)) from exc
        except UnknownFieldError as exc:
            raise UnknownFieldError(str(exc)) from exc
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc
        except KeyError as exc:
            raise UnknownFieldError(str(exc)) from exc

    # -------- Utilities / ergonomics --------
    def init(
        self,
        *,
        scope: Literal[
            "user", "user-local", "project", "project-local", "environment"
        ] = "user",
    ) -> None:
        scope = self._normalize_scope(scope)
        if scope == "env":
            return
        try:
            self._manager().init(scope)
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc

    def export_spec(self, dest: str | Path | None = None) -> Path:
        if dest is None:
            dest = Path(f"{self.provider_id}-spec.json")
        try:
            return _ORCH.export_spec(self.provider_id, dest)
        except PolicyError as exc:
            raise PolicyError(str(exc)) from exc

    def reload_spec(self) -> ProviderInfo:
        spec = _ORCH.reload_spec(self.provider_id)
        return _provider_info(spec)


    def target_path(
        self,
        scope: Literal[
            "user", "user-local", "project", "project-local", "environment"
        ] = "user",
    ) -> Path:

        """Return the file path used for *scope* writes."""
        scope = self._normalize_scope(scope)
        if scope == "env":
            raise PolicyError("Environment scope has no path")
        return policy.path(scope, self.provider_id)



import os
import pytest

from pysigil import api
from pysigil.orchestrator import Orchestrator
from pysigil.settings_metadata import IniFileBackend, IniSpecBackend
from tests.utils import DummyPolicy


def make_api(tmp_path):
    spec = IniSpecBackend(user_dir=tmp_path / "meta")
    policy = DummyPolicy(tmp_path / "user", tmp_path / "proj", host="host")
    cfg = IniFileBackend(policy=policy)
    orch = Orchestrator(spec_backend=spec, config_backend=cfg)
    api._ORCH = orch  # type: ignore[attr-defined]
    return api


def test_register_and_set(tmp_path):
    a = make_api(tmp_path)
    info = a.register_provider("my-pkg", title="My Package")
    assert info.provider_id == "my-pkg"
    # idempotent
    info2 = a.register_provider("my-pkg")
    assert info2.provider_id == "my-pkg"
    h = a.handle("my-pkg")
    h.add_field("retries", "integer", options={"minimum": 0})
    h.set("retries", 5)
    val = h.get("retries")
    assert val.value == 5
    assert val.source == "user"
    assert h.fields()[0].options == {"minimum": 0}
    assert "my-pkg" in a.providers()


def test_unknown_provider(tmp_path):
    a = make_api(tmp_path)
    with pytest.raises(a.UnknownProviderError):
        a.handle("missing")


def test_environment_scope(tmp_path, monkeypatch):
    a = make_api(tmp_path)
    a.register_provider("pkg")
    h = a.handle("pkg")
    h.add_field("api_field", "string")
    h.set("api_field", "42", scope="environment")
    assert os.environ["SIGIL_PKG_API_FIELD"] == "42"
    val = h.get("api_field")
    assert val.value == "42"
    assert val.source == "env"
    h.clear("api_field", scope="environment")
    assert "SIGIL_PKG_API_FIELD" not in os.environ
    val2 = h.get("api_field")
    assert val2.value is None
    assert val2.source is None
